/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.internal;

import java.net.InetSocketAddress;
import java.util.HashMap;
import java.util.LinkedList;

import org.apache.commons.lang3.tuple.Triple;

import com.aquenos.epics.jackie.client.DefaultChannelAccessClient;
import com.aquenos.epics.jackie.client.resolver.ChannelNameResolver;
import com.aquenos.epics.jackie.client.resolver.ChannelNameResolver.ChannelNameSearchResultProcessor;
import com.aquenos.epics.jackie.common.exception.ChannelAccessException;
import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessEventMask;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessStatus;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;
import com.aquenos.epics.jackie.common.util.IntegerIdPool;
import com.aquenos.epics.jackie.common.util.LinkedWeakIdentityHashSet;
import com.aquenos.epics.jackie.common.value.ChannelAccessGettableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessPuttableValue;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * <p>
 * Internal representation of a Channel Access channel. This class is only
 * intended for use by the classes inside this package and by the
 * {@link DefaultChannelAccessClient}.
 * </p>
 * 
 * <p>
 * Only one instance of this class is created for each channel name. An instance
 * of this class lives as long as there is at least one active
 * {@link ChannelAccessChannelImpl} referring to it.
 * </p>
 * 
 * <p>
 * This implementation takes care of finding the channel in the network and
 * connecting with the right server. It also takes care of starting the search
 * again when the connection with the server is lost.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class Channel {

    /**
     * Connection state of the channel.
     * 
     * @author Sebastian Marsching
     */
    public enum State {

        /**
         * Channel has just been created but not been started yet.
         */
        NEW,

        /**
         * Looking for the channel in the network.
         */
        SEARCHING,

        /**
         * Server has been found, but the registration of the channel with the
         * server has not completed yet.
         */
        CONNECTING,

        /**
         * Channel is connected and the server is responsive.
         */
        CONNECTED,

        /**
         * Channel is connected, but the server is unresponsive.
         */
        UNRESPONSIVE,

        /**
         * Channel has been destroyed and cannot be used any longer.
         */
        DESTROYED

    }

    private static final long SEARCH_DELAY = 30000L;

    private final Object lock = new Object();

    private boolean gotFirstUser;
    private LinkedWeakIdentityHashSet<ChannelAccessChannelImpl> channelFacades = new LinkedWeakIdentityHashSet<ChannelAccessChannelImpl>();
    private ChannelPool channelPool;
    private IntegerIdPool.Id cid;
    private boolean mayRead;
    private boolean mayWrite;
    private String name;
    private ChannelAccessValueType nativeDataType;
    private int nativeCount;
    private ChannelNameResolver resolver;
    private ChannelNameSearchResultProcessor searchResultProcessor = new ChannelNameSearchResultProcessor() {
        @Override
        public void processChannelLocation(String channelName, int cid,
                InetSocketAddress serverAddress,
                ChannelAccessVersion serverVersion) {
            synchronized (lock) {
                // We have to check whether we are still in the searching state,
                // because the channel might have been shutdown but the search
                // is cancelled asynchronously, so this method might still be
                // called.
                if (!state.equals(State.SEARCHING)) {
                    return;
                }
                if (!serverVersion.includes(ChannelAccessVersion.V4_4)) {
                    // We delay the next search request in order to avoid
                    // sending too many requests. We still want to keep
                    // searching for a channel because eventually, it might come
                    // up on a different server that supports Channel Access
                    // version 4.4.
                    search(30000L);
                } else {
                    connect(serverAddress, serverVersion);
                }
            }
        }

    };
    private Server server = null;
    private ServerPool serverPool;
    private int sid = 0;
    private State state = State.NEW;
    private IntegerIdPool subscriptionIdPool;
    private HashMap<Triple<Integer, ChannelAccessValueType, ChannelAccessEventMask>, Subscription> subscriptions = new HashMap<Triple<Integer, ChannelAccessValueType, ChannelAccessEventMask>, Subscription>();
    private LinkedWeakIdentityHashSet<Subscription> subscriptionsPendingCancellation = new LinkedWeakIdentityHashSet<Subscription>();

    /**
     * Creates a new channel. The channel starts in the {@link State#NEW} state.
     * Its {@link #start()} method has to be called in order to start it (start
     * the search for the channel in the network).
     * 
     * @param channelPool
     *            channel pool that is creating this channel. The channel will
     *            notify this channel pool when it is destroyed.
     * @param cidPool
     *            CID pool. This is the pool that is used to acquire a CID for
     *            this channel, so that all channel's of a client use a unique
     *            CID.
     * @param name
     *            name identifying the channel in the network.
     * @param resolver
     *            channel-name resolver that should be used for resolving the
     *            channel (finding the server for the channel in the network).
     * @param serverPool
     *            server pool. This server pool is used to get a connection to
     *            the server ({@link Server} instance) once the channel has been
     *            found.
     * @param subscriptionIdPool
     *            subscription ID pool. This is the pool that is used to acquire
     *            a unique ID for a subscription when one is created. This pool
     *            is shared among all servers / connections, so that a
     *            subscription can keep its ID when a channel moves to a
     *            different server.
     */
    public Channel(ChannelPool channelPool, IntegerIdPool cidPool, String name,
            ChannelNameResolver resolver, ServerPool serverPool,
            IntegerIdPool subscriptionIdPool) {
        this.channelPool = channelPool;
        this.cid = cidPool.acquireId();
        this.name = name;
        this.resolver = resolver;
        this.serverPool = serverPool;
        this.subscriptionIdPool = subscriptionIdPool;
    }

    /**
     * Returns the CID for this channel. The CID is a numeric ID uniquely
     * identifying the channel within the client. This ID is primarily used when
     * searching for a channel or connecting to a channel.
     * 
     * @return CID for this channel.
     * @throws IllegalStateException
     *             if this channel has already been destroyed and its CID has
     *             been released.
     */
    public int getCID() {
        return cid.get();
    }

    /**
     * Returns the name of this channel. This is the name that is used in search
     * requests for the channel.
     * 
     * @return name identifying this channel in the network.
     */
    public String getName() {
        return name;
    }

    /**
     * Returns the server that is currently providing this channel. Returns
     * <code>null</code> if the channel is current not connected and not
     * connecting.
     * 
     * @return server for this channel or <code>null</code>.
     */
    public Server getServer() {
        synchronized (lock) {
            return server;
        }
    }

    /**
     * Returns the channel's current connection state.
     * 
     * @return current connection state of this channel.
     */
    public State getState() {
        synchronized (lock) {
            return state;
        }
    }

    /**
     * Returns the read permission flag of the channel. This flag can only be
     * read if the channel is currently connected.
     * 
     * @return <code>true</code> if this channel is readable (get and monitor
     *         operations are allowed), <code>false</code> if such operations
     *         are not allowed.
     * @throws IllegalStateException
     *             if this channel is currently not connected.
     */
    public boolean isMayRead() {
        synchronized (lock) {
            if (!state.equals(State.CONNECTED)
                    && !state.equals(State.UNRESPONSIVE)) {
                throw new IllegalStateException("Channel is not connected.");
            }
            return mayRead;
        }
    }

    /**
     * Returns the write permission flag of the channel. This flag can only be
     * read if the channel is currently connected.
     * 
     * @return <code>true</code> if this channel is writable (put operations are
     *         allowed), <code>false</code> if such operations are not allowed.
     * @throws IllegalStateException
     *             if this channel is currently not connected.
     */
    public boolean isMayWrite() {
        synchronized (lock) {
            if (!state.equals(State.CONNECTED)
                    && !state.equals(State.UNRESPONSIVE)) {
                throw new IllegalStateException("Channel is not connected.");
            }
            return mayWrite;
        }
    }

    /**
     * Returns this channel's native count. This is the number of elements that
     * the process variable on the server has. Might be negative if the number
     * of elements on the server exceeds {@link Integer#MAX_VALUE}. This
     * information can only be read if the channel is currently connected.
     * 
     * @return number of elements the channel has on the server.
     * @throws IllegalStateException
     *             if this channel is currently not connected.
     */
    public int getNativeCount() {
        synchronized (lock) {
            if (!state.equals(State.CONNECTED)
                    && !state.equals(State.UNRESPONSIVE)) {
                throw new IllegalStateException("Channel is not connected.");
            }
            return nativeCount;
        }
    }

    /**
     * Returns this channel's native data-type. This is the data type that is
     * used for the process variable on the server and that the server can send
     * without having to perform any conversion. This information can only be
     * read if the channel is currently connected.
     * 
     * @return data-type of the channel that the server uses internally.
     * @throws IllegalStateException
     *             if this channel is currently not connected.
     */
    public ChannelAccessValueType getNativeDataType() {
        synchronized (lock) {
            if (!state.equals(State.CONNECTED)
                    && !state.equals(State.UNRESPONSIVE)) {
                throw new IllegalStateException("Channel is not connected.");
            }
            return nativeDataType;
        }
    }

    /**
     * <p>
     * Registers a channel facade with this channel. The channel facade is the
     * object that the user code sees as the channel instance. Internally, it
     * delegates all actions to this channel.
     * </p>
     * 
     * <p>
     * The list of registered channel facades is internally kept through weak
     * references. This allows the channel facades to be garbage collected when
     * they are not used any longer. When this list becomes empty (after at
     * least one channel facaded had been registered with this channel), this
     * channel is destroyed.
     * </p>
     * 
     * @param channelFacade
     *            channel facade that is going to use this channel.
     * @throws IllegalStateException
     *             if this channel has been destroyed.
     * @see #unregisterChannelFacade(ChannelAccessChannelImpl)
     */
    public void registerChannelFacade(ChannelAccessChannelImpl channelFacade) {
        synchronized (lock) {
            if (state.equals(State.DESTROYED)) {
                // We do not allow incrementing the use count on an already
                // destroyed channel, because it cannot be used anyway.
                throw new IllegalStateException(
                        "This channel is already destroyed.");
            }
            boolean wasNotInSetYet = channelFacades.add(channelFacade);
            // Each channel facade should only be registered once. We use an
            // assertion here because this method should only be called by
            // classes specific to this implementations.
            assert wasNotInSetYet;
            // We set a marker to indicate that there has been at least one
            // user. This means that the channel should be destroyed if the use
            // count goes back to zero.
            gotFirstUser = true;
        }
    }

    /**
     * <p>
     * Unregisters a channel facade from this channel. The channel facade is the
     * object that the user code sees as the channel instance. Internally, it
     * delegates all actions to this channel.
     * </p>
     * 
     * <p>
     * The list of registered channel facades is internally kept through weak
     * references. Even though this means that this references will eventually
     * be cleaned up if the channel facade is destroyed, we still unregister
     * channel facades explicitly to allow for an early release of resources.
     * This is also important to make sure that the subscription is destroyed
     * when there are no more monitor facades registered to it.
     * </p>
     * 
     * @param channelFacade
     *            channel facade that had been registered with this channel and
     *            is not using this channel any longer.
     * @see #registerChannelFacade(ChannelAccessChannelImpl)
     */
    public void unregisterChannelFacade(ChannelAccessChannelImpl channelFacade) {
        synchronized (lock) {
            boolean wasInSet = channelFacades.remove(channelFacade);
            // Each channel facade should only be unregistered once. We use an
            // assertion here because this method should only be called by
            // classes specific to this implementations.
            assert wasInSet;
            if (channelFacades.isEmpty()) {
                // This channel is not used any longer, so it should be
                // destroyed unless it is still connecting. If it is still
                // connecting we have to destroy it once it is connected
                // because we need to know the SID for the disconnect message.
                // We can also ignore the new state because this can only happen
                // if a second thread registered and unregistered from this
                // channel before the first thread that created this channel had
                // a chance to start it.
                if (state.equals(State.SEARCHING)
                        || state.equals(State.CONNECTED)
                        || state.equals(State.UNRESPONSIVE)) {
                    destroy();
                }
            }
        }
    }

    /**
     * Starts a read (get) operation. Such an operation can only be started if
     * the channel is connected. If the channel is disconnected before the
     * operation completes, the operation fails.
     * 
     * @param dataType
     *            data type to request from the server. If <code>null</code>,
     *            the native data-type (see {@link #getNativeDataType()}) is
     *            used.
     * @param count
     *            number of elements to request from the server. If zero or
     *            greater than the native count, the native count (see
     *            {@link #getNativeCount()}) is used.
     * @return future that is going to provide the result of the read operation.
     * @throws ChannelAccessException
     *             if the read operation cannot be started (e.g. because the
     *             channel is not connected or the data type and count would
     *             result in too large responses).
     */
    public ReadOperation<? extends ChannelAccessGettableValue<?>> readValue(
            ChannelAccessValueType dataType, int count) {
        synchronized (lock) {
            if (!state.equals(State.CONNECTED)
                    && !state.equals(State.UNRESPONSIVE)) {
                throw new ChannelAccessException(
                        ChannelAccessStatus.StatusMessage.ECA_DISCONN
                                .toStatus());
            }
            if (!mayRead) {
                throw new ChannelAccessException(
                        ChannelAccessStatus.StatusMessage.ECA_NORDACCESS
                                .toStatus());
            }
            if (dataType == null) {
                dataType = nativeDataType;
            }
            // The specified count should positive or zero. This is verified by
            // the calling code so that we only use an assertion here.
            assert (count >= 0);
            // A count of zero is interpreted as the native count. If the
            // specified count is greater than the native count we limit the
            // request to the native count. If we did not, the server would just
            // fill the elements that it does not have with zeros which would
            // simply waste bandwidth and bring no benefit at all.
            if (count == 0) {
                count = nativeCount;
                // The native count can be less than zero because (in theory) it
                // can exceed Integer.MAX_VALUE. In practice, such a high
                // element count is very unlikely to ever appear. We simply
                // limit to Integer.MAX_VALUE because we cannot store more
                // elements in a Java array anyway.
                if (count < 0) {
                    count = Integer.MAX_VALUE;
                }
            } else {
                if (count > nativeCount) {
                    count = nativeCount;
                }
            }
            // We have to call readChannel(...) while holding the mutex.
            // Clearly, this is not optimal because that method will acquire a
            // different mutex, but if we did not hold the mutex, the server
            // could change or become null before we get to call the method.
            // There should still be no risk of a dead lock because the server
            // will never call any code acquiring this object's mutex while it
            // holds its own mutex.
            return server.readChannel(this, dataType, count);
        }
    }

    /**
     * Starts a write (put) operation. Such an operation can only be started if
     * the channel is connected. If the channel is disconnected before the
     * operation completes, the operation fails.
     * 
     * @param value
     *            value that shall be written to the server.
     * @return future that is going to provide the result of the write
     *         operation. As a write operation does not have a resulting value,
     *         it only provides an indication that the operation finished or an
     *         exception if the operation failed.
     * @throws ChannelAccessException
     *             if the read operation cannot be started (e.g. because the
     *             channel is not connected or the data type and count would
     *             result in too large responses).
     */
    public WriteOperation writeValue(ChannelAccessPuttableValue<?> value) {
        synchronized (lock) {
            if (!state.equals(State.CONNECTED)
                    && !state.equals(State.UNRESPONSIVE)) {
                throw new ChannelAccessException(
                        ChannelAccessStatus.StatusMessage.ECA_DISCONN
                                .toStatus());
            }
            if (!mayWrite) {
                throw new ChannelAccessException(
                        ChannelAccessStatus.StatusMessage.ECA_NOWTACCESS
                                .toStatus());
            }
            if (value.getValueSize() == 0
                    || (value.getValueSize() > nativeCount && nativeCount >= 0)) {
                throw new ChannelAccessException(
                        ChannelAccessStatus.StatusMessage.ECA_BADCOUNT
                                .toStatus());
            }
            // We have to call writeChannel(...) while holding the mutex.
            // Clearly, this is not optimal because that method will acquire a
            // different mutex, but if we did not hold the mutex, the server
            // could change or become null before we get to call the method.
            // There should still be no risk of a dead lock because the server
            // will never call any code acquiring this object's mutex while it
            // holds its own mutex.
            return server.writeChannel(this, value);
        }
    }

    /**
     * Returns a subscription for a monitor with the specified parameters. If no
     * such subscription exists, a new one is created.
     * 
     * @param count
     *            maximum number of elements that should be sent by the server
     *            or zero to use the channel's native count (always receive all
     *            elements).
     * @param dataType
     *            data type to request from the server or <code>null</code> to
     *            use the channel's native data-type.
     * @param mask
     *            event mask to be used. This defines when the server sends an
     *            event. Must not be <code>DBE_NONE</code>.
     * @return subscription for the specified parameters. The returned
     *         subscription might be destroyed asynchronously, so the calling
     *         code has to expect that it has to deal with a destroyed
     *         subscription. It should call
     *         {@link #removeSubscription(Subscription)} to remove the
     *         subscription an then request a new one.
     */
    public Subscription findOrCreateSubscription(int count,
            ChannelAccessValueType dataType, ChannelAccessEventMask mask) {
        Subscription subscription;
        Triple<Integer, ChannelAccessValueType, ChannelAccessEventMask> subscriptionKey = Triple
                .of(count, dataType, mask);
        synchronized (lock) {
            if (state.equals(State.DESTROYED)) {
                // We do not allow creating a subscription for a destroyed
                // channel because it cannot be used anyway.
                throw new IllegalStateException(
                        "This channel is already destroyed.");
            }
            subscription = subscriptions.get(subscriptionKey);
            if (subscription == null) {
                subscription = new Subscription(this, count, dataType, mask,
                        subscriptionIdPool.acquireId());
                subscriptions.put(subscriptionKey, subscription);
                // If the channel is connected, we register the subscription
                // with the server. Otherwise, it will be registered when the
                // channel connects.
                if (state.equals(State.CONNECTED)
                        || state.equals(State.UNRESPONSIVE)) {
                    ChannelAccessValueType actualDataType = dataType;
                    if (actualDataType == null) {
                        actualDataType = nativeDataType;
                    }
                    int actualCount = count;
                    if (actualCount == 0) {
                        actualCount = nativeCount;
                    }
                    if (mayRead) {
                        try {
                            server.registerSubscription(subscription,
                                    actualDataType, actualCount);
                        } catch (ChannelAccessException e) {
                            // In this special case it is okay to call the a
                            // notification method of the subscription while
                            // holding the mutex. As we have just created the
                            // subscription, we know that no one else can be
                            // using it and thus there is no risk of a dead
                            // lock. We ignore an ECA_DISCONN error because the
                            // subscription is connected when the channel
                            // connects again.
                            if (!e.getStatus()
                                    .matchesMessage(
                                            ChannelAccessStatus.StatusMessage.ECA_DISCONN)) {
                                subscription.errorReceived(server,
                                        e.getStatus(), e.getExtendedMessage());
                            }
                        }
                    } else {
                        // In this special case it is okay to call the a
                        // notification method of the subscription while holding
                        // the mutex. As we have just created the subscription,
                        // we know that no one else can be using it and thus
                        // there is no risk of a dead lock.
                        subscription
                                .errorReceived(
                                        server,
                                        ChannelAccessStatus.StatusMessage.ECA_NORDACCESS
                                                .toStatus(), null);
                    }
                }
            }
            // We might return a destroyed subscription. The calling code has to
            // check this and request a new one if needed.
            return subscription;
        }
    }

    /**
     * <p>
     * Removes the subscription from the channel. This is called by the
     * subscription when it is destroyed, but it can also be called by a channel
     * facade when it encounters a destroyed subscription and need to get a new
     * one.
     * </p>
     * 
     * <p>
     * Removing the subscription has the effect of request it to be removed from
     * the server as well. When the subscription has finally been removed, its
     * {@link Subscription#cancelled(Server) cancelled(...)} method is called.
     * If the server disconnects (or is already disconnected) the subscription's
     * {@link Subscription#serverDisconnected() serverDisconnected()} method is
     * called.
     * </p>
     * 
     * @param subscription
     *            subscription to be removed from this channel.
     */
    public void removeSubscription(Subscription subscription) {
        boolean connected;
        synchronized (lock) {
            Triple<Integer, ChannelAccessValueType, ChannelAccessEventMask> subscriptionKey = Triple
                    .of(subscription.getCount(), subscription.getDataType(),
                            subscription.getMask());
            Subscription subscriptionRemoved = subscriptions
                    .remove(subscriptionKey);
            // There is a small chance that a different subscription with the
            // same settings has been registered before the old subscription
            // removed itself. In this case, we want to add the new subscription
            // again. We can also return immediately, because all the other
            // actions must already have been triggered, when the old
            // subscription was removed.
            if (subscriptionRemoved != subscription) {
                subscriptions.put(subscriptionKey, subscriptionRemoved);
                return;
            }
            // If we are currently not connected, we do not have to do anything
            // else because effectively, the subscription is currently not
            // registered with any server. However, we have to notify the
            // subscription so that it knows that it can release its ID.
            connected = (state == State.CONNECTED || state == State.UNRESPONSIVE);
            if (connected) {
                // We have to remember that the subscription is pending removal
                // so that we can notify it when we lose the connection to the
                // server before it sends the cancellation message.
                subscriptionsPendingCancellation.add(subscription);
                ChannelAccessValueType dataType = subscription.getDataType();
                if (dataType == null) {
                    dataType = nativeDataType;
                }
                int count = subscription.getCount();
                if (count == 0) {
                    count = nativeCount;
                }
                // We can call unregisterSubscription because even though it
                // acquires a mutex, that mutex and the mutex hold by this
                // method are always acquired in the same order.
                // There is a race condition that might cause an
                // IllegalStateException when the server has disconnected but
                // this channel is still in the connected state. This can happen
                // because when the server is disconnected, the server's state
                // changes first and its channels are notified after releasing
                // the mutex that protects the state from changing. This is
                // necessary in order to avoid a dead lock. If we catch such an
                // exception, we know that the server has been disconnected and
                // we can treat this like if the channel had already been
                // notified.
                try {
                    server.unregisterSubscription(subscription, dataType, count);
                } catch (IllegalStateException e) {
                    // We have to set the connected flag to false so that the
                    // subscription is notified and can release its ID.
                    connected = false;
                }
            }
        }
        // If we are not connected, we have to notify the subscription about
        // this, so that it can release its ID.
        if (!connected) {
            subscription.serverDisconnected();
        }
    }

    /**
     * Notifies the channel that a subscription has finally been cancelled. This
     * always happens after it has been removed, when the subscription is
     * finally notified by the server that it has removed the subscription.
     * 
     * @param subscription
     *            subscription that has been cancelled by the server.
     */
    public void subscriptionCancelled(Subscription subscription) {
        // We remove the subscription explicitly. This should not be necessary
        // because we only hold a weak reference to the subscription, but it is
        // still better to be explicit about cleaning up the resources.
        synchronized (lock) {
            subscriptionsPendingCancellation.remove(subscription);
        }
    }

    /**
     * <p>
     * Starts this channel. This is separated from the constructor so that the
     * {@link ChannelPool} can first check whether there has been a concurrent
     * request to create a channel before starting a newly created channel.
     * </p>
     * 
     * <p>
     * Starting a channel has the effect that a search request is queued with
     * the {@link ChannelNameResolver}.
     * </p>
     * 
     * @throws IllegalStateException
     *             if the channel name resolver has been destroyed. Typically,
     *             this only happens if the client is shutting down.
     * @throws SerializedChannelNameTooLargeException
     *             if the serialized form of the channel name is too long to be
     *             sent as part of a search request.
     */
    public void start() {
        synchronized (lock) {
            if (!state.equals(State.NEW)) {
                return;
            }
            boolean success = false;
            try {
                // Calling search may result in an IllegalStateException or a
                // SerializedChannelNameTooLargeException.
                search(0L);
                success = true;
            } finally {
                if (!success) {
                    state = State.DESTROYED;
                    cid.release();
                } else {
                    state = State.SEARCHING;
                }
            }
        }
    }

    /**
     * Releases this channel. This is very similar to {@link #destroy()},
     * however it has to be used when the channel has not been started yet.
     * Calling this method has the effect of setting the channel's state to
     * {@link State#DESTROYED DESTROYED} and releasing the CID.
     * 
     * @throws IllegalStateException
     *             if the channel is not in the {@link State#NEW NEW} state.
     */
    public void release() {
        synchronized (lock) {
            if (!state.equals(State.NEW)) {
                // This method should only be called if the channel has not
                // been started yet.
                throw new IllegalStateException(
                        "The channel has already been started.");
            }
            state = State.DESTROYED;
            cid.release();
        }
    }

    private void connect(InetSocketAddress serverAddress,
            ChannelAccessVersion serverVersion) {
        // This method is only called while holding the mutex and after checking
        // the current state.
        assert (Thread.holdsLock(lock));
        assert (state.equals(State.SEARCHING));
        state = State.CONNECTING;
        // We try to register the channel until we get a server that has not
        // been destroyed yet or we hit another kind of problem.
        for (;;) {
            this.server = serverPool.findOrCreateServer(serverAddress,
                    serverVersion);
            // It is okay to call registerChannel while holding the mutex. We
            // have to hold the mutex in order to avoid a race condition and
            // even though registerChannel acquires a different mutex, the
            // mutexes will always be acquired in the same order (e.g. when
            // unregistering from the server), so there is no risk of a dead
            // lock.
            try {
                this.server.registerChannel(this);
                break;
            } catch (IllegalStateException e) {
                // An IllegalStateException might be thrown because the server
                // has been destroyed by another thread before we got a chance
                // to register the channel. In this case we simply request a new
                // server until we get one that has not been destroyed yet.
                if (this.server
                        .getState()
                        .equals(ChannelAccessPVServerConnection.State.CONNECTION_DESTROYED)) {
                    continue;
                } else {
                    // If something else caused the exception, we throw it
                    // again.
                    throw e;
                }
            }
        }
    }

    private void destroy() {
        // This method is only called while holding the mutex and after checking
        // the current state.
        assert (Thread.holdsLock(lock));
        if (state.equals(State.CONNECTED) || state.equals(State.UNRESPONSIVE)) {
            // Calling unregisterChannel while holding the mutex is not
            // optimal because that method will acquire a different mutex.
            // However, we need to hold the mutex because otherwise another
            // channel object for the same channel might be created which
            // could cause registerChannel(...) and unregisterChannel(...)
            // to be called in the wrong order. There should still be no
            // risk of a dead lock because the server will never call any
            // code acquiring this object's mutex while it holds its own
            // mutex.
            server.unregisterChannel(this);
        } else if (state.equals(State.SEARCHING)) {
            // We have to cancel the search for the channel while holding
            // the mutex because otherwise we might cancel the search after
            // the channel has been revived if the channel is destroyed and
            // revived in a short interval.
            resolver.cancelChannelName(cid.get());
        } else {
            // This method should never be called in the new, connecting, or
            // destroyed state.
            throw new AssertionError("destroy() has been called in state "
                    + state + ".");
        }
        state = State.DESTROYED;
        server = null;
        cid.release();
        // Remove this channel from the ChannelPool to avoid accumulating unused
        // objects. It is save to do this while holding the mutex because the
        // map used for the channel pool is lock free.
        channelPool.removeChannel(this);
    }

    private void destroyConnecting() {
        // This version of the destroy method should be used when the channel is
        // in the connecting state but can safely be destroyed because the
        // connection failed. This method is only called while holding the mutex
        // and after checking the current state.
        assert (Thread.holdsLock(lock));
        assert (state.equals(State.CONNECTING));
        state = State.DESTROYED;
        cid.release();
        // Remove this channel from the ChannelPool to avoid accumulating unused
        // objects. It is save to do this while holding the mutex because the
        // map used for the channel pool is lock free.
        channelPool.removeChannel(this);
    }

    private void search(long delay) {
        // This method is only called while holding the mutex and after checking
        // the current state.
        assert (Thread.holdsLock(lock));
        assert (!state.equals(State.DESTROYED));
        try {
            // Calling resolveChannelName(...) while holding the mutex should
            // be safe because resolveChannelName should be lock free.
            resolver.resolveChannelName(name, cid.get(), searchResultProcessor,
                    delay);
            state = State.SEARCHING;
        } catch (IllegalStateException e) {
            // If the channel name resolver has already been destroyed, we
            // cannot add a search request. This means that the client must
            // be in shutdown and we simply destroy this channel as well.
            // We do not use the destroy() method for that, as the logic is
            // slightly different for this special case.
            state = State.DESTROYED;
            cid.release();
            // We rethrow the exception so that the calling code does not get
            // stuck in an infinite loop trying to create the channel because
            // it believes the channel was destroyed by a different thread.
            throw e;
        }
    }

    private void disconnectInternal(Server server) {
        LinkedList<ChannelAccessChannelImpl> listeners;
        LinkedList<Subscription> subscriptionsPendingCancellation;
        synchronized (lock) {
            // If the notification is from a different server, we ignore it.
            // This might happen because notifications may be received after
            // unregistering a channel.
            if (this.server != server) {
                return;
            }
            this.server = null;
            if (state.equals(State.CONNECTING)) {
                // The channel might have lost all of its users before it got
                // connected. In this case, we have to destroy it right away. We
                // cannot do this before the channel is connected (or the
                // connection failed) because we normally need the SID. If there
                // are no users, but the got-first-user flag has not been set
                // yet, it just means that the connection attempt failed more
                // quickly than the thread that created this channel could
                // register its facade. In this case we obviously do not want to
                // destroy this channel.
                if (channelFacades.isEmpty() && gotFirstUser) {
                    destroyConnecting();
                } else {
                    try {
                        search(0L);
                    } catch (Throwable e) {
                        // search(...) only throws if there is a problem with
                        // the channel name or if the resolver has been
                        // destroyed. If there was a problem with the channel
                        // name, we would have gotten the exception when we
                        // created the channel, so the resolver must have been
                        // destroyed which means that we are in shutdown. In
                        // this case, we simply destroy this channel and ignore
                        // the exception.
                        destroyConnecting();
                    }
                }
                // If we are in the connecting state, we do not want to notify
                // the listeners.
                return;
            } else if (state.equals(State.UNRESPONSIVE)
                    || state.equals(State.CONNECTED)) {
                // When we got disconnected, we can notify all subscriptions
                // that have been waiting for the confirmation of their
                // cancellation.
                subscriptionsPendingCancellation = new LinkedList<Subscription>(
                        this.subscriptionsPendingCancellation);
                this.subscriptionsPendingCancellation.clear();
                try {
                    search(0L);
                } catch (Throwable e) {
                    // search(...) only throws if there is a problem with
                    // the channel name or if the resolver has been
                    // destroyed. If there was a problem with the channel
                    // name, we would have gotten the exception when we
                    // created the channel, so the resolver must have been
                    // destroyed which means that we are in shutdown. In
                    // this case, we simply destroy this channel and ignore
                    // the exception.
                    destroy();
                }
            } else {
                // If we are in any other state, we do not want to notify the
                // listeners.
                return;
            }
            // We have to get the list of listeners while we are still holding
            // the lock. This ensures that the list is not modified concurrently
            // It also ensures that the listening channel facade is notified
            // about every state change after it registered.
            listeners = new LinkedList<ChannelAccessChannelImpl>(channelFacades);
        }
        // We notify the listeners after releasing the mutex in order to avoid
        // dead locks. This is okay because notifications of state changes are
        // only processed in the communication thread and we change the state
        // attribute while holding the mutex. The worst thing that can happen is
        // that one of the listeners is notified even though it already
        // unregistered or that it is notified of a state change even though it
        // was already aware of this state. Both cases are handled correctly by
        // the listener.
        for (ChannelAccessChannelImpl listener : listeners) {
            // channelDisconnected() does not throw (it catches exceptions from
            // the listeners registered to the facade).
            listener.channelDisconnected();
        }
        for (Subscription subscription : subscriptionsPendingCancellation) {
            subscription.serverDisconnected();
        }
    }

    /**
     * <p>
     * Returns the SID for this channel. The SID is a server-assigned ID that
     * uniquely identifies a channel within a server.
     * </p>
     * 
     * <p>
     * This method may only be called by the {@link Server} instance belonging
     * to this channel and only while it has been called from a method of this
     * channel that holds this channel's mutex.
     * </p>
     * 
     * <p>
     * This design is rather ugly, but there is no good way around it: As the
     * SID changes when the channel connects with a different server, access to
     * it must be protected by a mutex. However, we cannot take the mutex in
     * this method because the channel mutex always must be taken before the
     * server mutex. If we allowed a different order, we might get dead locks.
     * </p>
     * 
     * @return SID currently assigned to this channel. Only valid if the channel
     *         is connected.
     */
    int getSID() {
        // We use assertions to check that we hold the mutex and are in the
        // connected state. If one of this assertions does not hold, some code
        // that really should not call this method called it.
        assert Thread.holdsLock(lock);
        assert (state.equals(State.CONNECTED) || state
                .equals(State.UNRESPONSIVE));
        return sid;
    }

    /**
     * Notifies the channel that the server it has been registered with has been
     * destroyed. A server is destroyed when the connection is lost. This method
     * may only be called by the {@link Server} associated with this channel.
     * 
     * @param server
     *            server that has been destroyed.
     */
    void serverDestroyed(Server server) {
        disconnectInternal(server);
    }

    /**
     * Notifies the channel that the server it has been registered with is now
     * responsive. This method may only be called by the {@link Server}
     * associated with this channel.
     * 
     * @param server
     *            server that changed to the responsive state.
     */
    void serverResponsive(Server server) {
        LinkedList<ChannelAccessChannelImpl> listeners;
        synchronized (lock) {
            // We only want to process notifications from our current server.
            // Performing this check makes sure that we are immune to race
            // conditions in the registration / unregistration with servers.
            if (server != this.server) {
                return;
            }
            // We are only interested in a notification about a responsive
            // server, if we have been in the unresponsive state.
            if (!this.state.equals(State.UNRESPONSIVE)) {
                return;
            }
            this.state = State.CONNECTED;
            // We have to get the list of listeners while we are still holding
            // the lock. This ensures that the list is not modified concurrently
            // It also ensures that the listening channel facade is notified
            // about every state change after it registered.
            listeners = new LinkedList<ChannelAccessChannelImpl>(channelFacades);
        }
        // We notify the listeners after releasing the mutex in order to avoid
        // dead locks. This is okay because notifications of state changes are
        // only processed in the communication thread and we change the state
        // attribute while holding the mutex. The worst thing that can happen is
        // that one of the listeners is notified even though it already
        // unregistered or that it is notified of a state change even though it
        // was already aware of this state. Both cases are handled correctly by
        // the listener.
        for (ChannelAccessChannelImpl listener : listeners) {
            // channelConnected() does not throw (it catches exceptions from the
            // listeners registered to the facade).
            listener.channelConnected();
        }
    }

    /**
     * Notifies the channel that the server it has been registered with is now
     * unresponsive. This method may only be called by the {@link Server}
     * associated with this channel.
     * 
     * @param server
     *            server that changed to the unresponsive state.
     */
    void serverUnresponsive(Server server) {
        LinkedList<ChannelAccessChannelImpl> listeners;
        synchronized (lock) {
            // We only want to process notifications from our current server.
            // Performing this check makes sure that we are immune to race
            // conditions in the registration / unregistration with servers.
            if (server != this.server) {
                return;
            }
            // We are only interested in a notification about an unresponsive
            // server, if we have been in the responsive state.
            if (!this.state.equals(State.CONNECTED)) {
                return;
            }
            this.state = State.UNRESPONSIVE;
            // We have to get the list of listeners while we are still holding
            // the lock. This ensures that the list is not modified concurrently
            // It also ensures that the listening channel facade is notified
            // about every state change after it registered.
            listeners = new LinkedList<ChannelAccessChannelImpl>(channelFacades);
        }
        // We notify the listeners after releasing the mutex in order to avoid
        // dead locks. This is okay because notifications of state changes are
        // only processed in the communication thread and we change the state
        // attribute while holding the mutex. The worst thing that can happen is
        // that one of the listeners is notified even though it already
        // unregistered or that it is notified of a state change even though it
        // was already aware of this state. Both cases are handled correctly by
        // the listener.
        for (ChannelAccessChannelImpl listener : listeners) {
            // channelDisconnected() does not throw (it catches exceptions from
            // the listeners registered to the facade).
            listener.channelDisconnected();
        }
    }

    /**
     * Notifies the channel that it has been connected on the server side. This
     * method may only be called by the {@link Server} associated with this
     * channel.
     * 
     * @param server
     *            server that this channel has been registered with.
     * @param sid
     *            server-assigned ID that identifies the channel.
     * @param nativeDataType
     *            native data-type of the channel. This is the data-type that
     *            the server uses and that requires no conversions on the server
     *            side.
     * @param nativeCount
     *            native count of the channel. This is the number of elements
     *            that the channel's value has on the server side.
     * @param accessRightsValid
     *            <code>true</code> if access rights for the channel have been
     *            received from the server. <code>false</code> if no access
     *            rights have been received.
     * @param mayRead
     *            <code>true</code> if the channel allows read operations,
     *            <code>false</code> otherwise. This value is undefined if
     *            <code>accessRightsValid</code> is <code>false</code> and
     *            should not be used in this case.
     * @param mayWrite
     *            <code>true</code> if the channel allows write operations,
     *            <code>false</code> otherwise. This value is undefined if
     *            <code>accessRightsValid</code> is <code>false</code> and
     *            should not be used in this case.
     */
    void connected(Server server, int sid,
            ChannelAccessValueType nativeDataType, int nativeCount,
            boolean accessRightsValid, boolean mayRead, boolean mayWrite) {
        LinkedList<ChannelAccessChannelImpl> listeners;
        LinkedList<Triple<Subscription, ChannelAccessStatus, String>> subscriptionErrors = new LinkedList<Triple<Subscription, ChannelAccessStatus, String>>();
        synchronized (lock) {
            // If the notification is from a different server, we ignore it.
            // This might happen because notifications may be received after
            // unregistering a channel.
            if (this.server != server) {
                return;
            }
            // By checking whether the notification is from the right server, we
            // implicitly checked that we are in the right state. Otherwise, the
            // server would not match.
            this.sid = sid;
            this.nativeDataType = nativeDataType;
            this.nativeCount = nativeCount;
            // If we did not receive any access rights, we assume that both
            // access rights are false (the C client does the same). For CA
            // versions prior to Channel Access version 4.1, we would have to
            // assume that the access rights are true, but we do not support
            // these old versions anyway (we require version 4.4 as a minimum).
            if (accessRightsValid) {
                this.mayRead = mayRead;
                this.mayWrite = mayWrite;
            } else {
                this.mayRead = false;
                this.mayWrite = false;
            }
            // If the channel got connected but the server is unresponsive, we
            // want to set the channel to the unresponsive state as well.
            if (this.server
                    .getState()
                    .equals(ChannelAccessPVServerConnection.State.CONNECTION_RESPONSIVE)) {
                state = State.CONNECTED;
            } else {
                state = State.UNRESPONSIVE;
            }
            // Now that the channel is connected, we can register the
            // subscriptions with the server.
            for (Subscription subscription : subscriptions.values()) {
                ChannelAccessValueType dataType = subscription.getDataType();
                if (dataType == null) {
                    dataType = nativeDataType;
                }
                int count = subscription.getCount();
                if (count == 0) {
                    count = nativeCount;
                }
                if (mayRead) {
                    try {
                        // We can call registerSubscription(...) while holding
                        // the mutex because even though that method uses a
                        // different mutex, the two mutexes are always acquired
                        // in the same order and thus there is no risk of a dead
                        // lock.
                        server.registerSubscription(subscription, dataType,
                                count);
                    } catch (ChannelAccessException e) {
                        // We cannot notify the subscription while holding the
                        // mutex, therefore we have to put it on a queue and
                        // notify it after releasing the mutex. We ignore an
                        // ECA_DISCONN error because the subscription is
                        // connected when the channel connects again.
                        if (!e.getStatus().matchesMessage(
                                ChannelAccessStatus.StatusMessage.ECA_DISCONN)) {
                            subscriptionErrors.add(Triple.of(subscription,
                                    e.getStatus(), e.getExtendedMessage()));
                        }
                    }
                } else {
                    // We cannot notify the subscription while holding the
                    // mutex, therefore we have to put it on a queue and
                    // notify it after releasing the mutex.
                    subscriptionErrors.add(Triple.of(subscription,
                            ChannelAccessStatus.StatusMessage.ECA_NORDACCESS
                                    .toStatus(), (String) null));
                }
            }
            // The channel might have lost all of its users before it got
            // connected. In this case, we have to destroy it right away. We
            // cannot do this before the channel is connected because we need
            // the SID. If there are no users, but the got-first-user flag has
            // not been set yet, it just means that the connection was
            // established more quickly than the the thread that created this
            // channel could register its facade. In this case we obviously
            // do not want to destroy this channel.
            if (channelFacades.isEmpty() && gotFirstUser) {
                destroy();
            }
            // We have to get the list of listeners while we are still holding
            // the lock. This ensures that the list is not modified concurrently
            // It also ensures that the listening channel facade is notified
            // about every state change after it registered. We do not want to
            // notify the listeners if the server (and thus the channel) is not
            // responsive.
            if (state.equals(State.CONNECTED)) {
                listeners = new LinkedList<ChannelAccessChannelImpl>(
                        channelFacades);
            } else {
                return;
            }
        }
        // We notify the listeners after releasing the mutex in order to avoid
        // dead locks. This is okay because notifications of state changes are
        // only processed in the communication thread and we change the state
        // attribute while holding the mutex. The worst thing that can happen is
        // that one of the listeners is notified even though it already
        // unregistered or that it is notified of a state change even though it
        // was already aware of this state. Both cases are handled correctly by
        // the listener.
        for (ChannelAccessChannelImpl listener : listeners) {
            // channelConnected() does not throw (it catches exceptions from the
            // listeners registered to the facade).
            listener.channelConnected();
        }
        for (Triple<Subscription, ChannelAccessStatus, String> subscriptionError : subscriptionErrors) {
            subscriptionError.getLeft()
                    .errorReceived(server, subscriptionError.getMiddle(),
                            subscriptionError.getRight());
        }
    }

    /**
     * Notifies the channel that the server has refused to connect the channel.
     * This means that the channel is still not connected. This method may only
     * be called by the {@link Server} associated with this channel.
     * 
     * @param server
     *            server that this channel has been registered with.
     */
    void connectFailed(Server server) {
        synchronized (lock) {
            // If the notification is from a different server, we ignore it.
            // This might happen because notifications may be received after
            // unregistering a channel.
            if (this.server != server) {
                return;
            }
            this.server = null;
            if (state.equals(State.CONNECTING)) {
                // The channel might have lost all of its users before it got
                // connected. In this case, we have to destroy it right away. We
                // cannot do this before the channel is connected because we
                // need the SID. If the use count is zero, but the
                // got-first-user flag has not been set yet, it just means that
                // the connection was established more quickly than the the
                // thread that created this channel could register its facade.
                // In this case we obviously do not want to destroy this
                // channel.
                if (channelFacades.isEmpty() && gotFirstUser) {
                    destroyConnecting();
                } else {
                    // We delay the next search in order to avoid overloading a
                    // server that does not allow the channel to be connected
                    // but advertises it.
                    search(SEARCH_DELAY);
                }
            } else {
                return;
            }
        }
    }

    /**
     * Notifies the channel that the server has decided to disconnect it. This
     * means that the channel is no longer connected. This method may only be
     * called by the {@link Server} associated with this channel.
     * 
     * @param server
     *            server that this channel has been registered with.
     */
    void disconnectedByServer(Server server) {
        disconnectInternal(server);
    }

}
