/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.internal;

import com.aquenos.epics.jackie.client.DefaultClientThreadingStrategy;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.io.CommunicationProcessor;

/**
 * <p>
 * Variant of a communication thread that only has a single communication
 * processor that is already known at construction time. This specialization
 * allows the communication processor's
 * {@link CommunicationProcessor#setCommunicationController(com.aquenos.epics.jackie.common.io.CommunicationController)
 * setCommunicationController(...)} method to be called from the communication
 * thread which has some benefits because it avoids unnecessary synchronization
 * between different threads.
 * </p>
 * 
 * <p>
 * This class is only intended for use by {@link DefaultClientThreadingStrategy}
 * .
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class SingleProcessorCommunicationThread extends CommunicationThread {

    private CommunicationProcessor communicationProcessor;

    /**
     * Creates a communication thread. The thread uses the specified name and is
     * created as a daemon thread. However, the thread is not started
     * automatically but has to be started by calling its {@link #start()}
     * method. After the thread has been started, the communication processor's
     * <code>setCommunicationController(...)</code> method is called
     * automatically.
     * 
     * @param threadName
     *            name of the new thread.
     * @param communicationProcessor
     *            communication processor that should be processed by the
     *            thread.
     * @param errorHandler
     *            error handler that is used to log errors that are thrown by
     *            the communication controller.
     */
    public SingleProcessorCommunicationThread(String threadName,
            CommunicationProcessor communicationProcessor,
            ErrorHandler errorHandler) {
        super(threadName, errorHandler);
        assert (communicationProcessor != null);
        this.communicationProcessor = communicationProcessor;
    }

    @Override
    public void run() {
        communicationProcessor
                .setCommunicationController(getCommunicationController());
        super.run();
    }

}
