/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver;

import java.net.InetSocketAddress;

import com.aquenos.epics.jackie.common.exception.SerializedChannelNameTooLargeException;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConstants;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;

/**
 * <p>
 * Resolves channel names to network locations. These network locations can then
 * be used to establish connections with the respective channels.
 * </p>
 * 
 * <p>
 * An implementation of this interface should ensure that the methods defined by
 * this interface are thread safe and thus can be called by any thread.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public interface ChannelNameResolver {

    /**
     * Processes the result of a channel name lookup.
     * 
     * @author Sebastian Marsching
     */
    public interface ChannelNameSearchResultProcessor {

        /**
         * Processes the results of a successful channel name lookup. This
         * method is only called if the lookup is successful. The resolver will
         * not continue trying to resolve the channel name after calling this
         * method. If the channel is never found, this method might never be
         * called. Typically, this method will be called from the communication
         * thread of the channel name resolver. Therefore, this method should
         * not perform operations that take a considerable amount of time or
         * might block.
         * 
         * @param channelName
         *            channel name that has been resolved (never
         *            <code>null</code>).
         * @param cid
         *            client-assigned numeric identifier for the channel. This
         *            is the same identifier that was passed to
         *            {@link ChannelNameResolver#resolveChannelName(String, int, ChannelNameSearchResultProcessor)}
         *            .
         * @param serverAddress
         *            IP address and TCP port of the Channel Access server that
         *            provides the channel (never <code>null</code>). This
         *            address can be used to create a socket for the channel.
         * @param serverVersion
         *            Channel Access version supported by the server (never
         *            <code>null</code>). This is the Channel Access version
         *            that should be used when connecting to the server.
         */
        void processChannelLocation(String channelName, int cid,
                InetSocketAddress serverAddress,
                ChannelAccessVersion serverVersion);

    }

    /**
     * Tries to resolve a channel name. When the channel name can be resolved,
     * the <code>resultProcessor</code> is called. When the search for the
     * channel fails with an exception, the <code>exceptionProcessor</code> is
     * called. If the channel can simply not be resolved, no call-back is called
     * and the resolver simply continues with trying to resolve the channel
     * name.
     * 
     * @param channelName
     *            name of the channel that shall be resolved (never
     *            <code>null</code>).
     * @param cid
     *            client-assigned numeric identifier for the channel. This
     *            identifier is used as a key in various places and thus must be
     *            unique. Using the same identifier for multiple search requests
     *            is an error and will result in unpredictable behavior. If the
     *            search for a channel has succeeded of failed finally (one of
     *            the two call-backs has been called), the CID can be reused.
     *            However, it is recommended not to reuse the CID for a certain
     *            amount of time because otherwise late replies for the
     *            preceding search requests might be misinterpreted as replies
     *            to the new search requests.
     * @param resultProcessor
     *            processor that is called when the search for the channel
     *            succeeds (never <code>null</code>).
     * @throws IllegalStateException
     *             if the channel name resolver has been destroyed.
     * @throws SerializedChannelNameTooLargeException
     *             if the channel name is too large to be serialized. This can
     *             happen because the serialized form of the channel name
     *             exceeds the general limit defined by
     *             {@link ChannelAccessConstants#MAX_CHANNEL_NAME_BYTES_LENGTH}
     *             or because it is too long too fit into a single UDP packet.
     */
    void resolveChannelName(String channelName, int cid,
            ChannelNameSearchResultProcessor resultProcessor);

    /**
     * Tries to resolve a channel name. When the channel name can be resolved,
     * the <code>resultProcessor</code> is called. When the search for the
     * channel fails with an exception, the <code>exceptionProcessor</code> is
     * called. If the channel can simply not be resolved, no call-back is called
     * and the resolver simply continues with trying to resolve the channel
     * name.
     * 
     * @param channelName
     *            name of the channel that shall be resolved (never
     *            <code>null</code>).
     * @param cid
     *            client-assigned numeric identifier for the channel. This
     *            identifier is used as a key in various places and thus must be
     *            unique. Using the same identifier for multiple search requests
     *            is an error and will result in unpredictable behavior. If the
     *            search for a channel has succeeded of failed finally (one of
     *            the two call-backs has been called), the CID can be reused.
     *            However, it is recommended not to reuse the CID for a certain
     *            amount of time because otherwise late replies for the
     *            preceding search requests might be misinterpreted as replies
     *            to the new search requests.
     * @param resultProcessor
     *            processor that is called when the search for the channel
     *            succeeds (never <code>null</code>).
     * @param initialDelay
     *            if greater than zero, the start of the search is delayed by
     *            the specified number of milliseconds. Otherwise, the search is
     *            started as soon as possible.
     * @throws IllegalStateException
     *             if the channel name resolver has been destroyed.
     * @throws SerializedChannelNameTooLargeException
     *             if the channel name is too large to be serialized. This can
     *             happen because the serialized form of the channel name
     *             exceeds the general limit defined by
     *             {@link ChannelAccessConstants#MAX_CHANNEL_NAME_BYTES_LENGTH}
     *             or because it is too long too fit into a single UDP packet.
     */
    void resolveChannelName(String channelName, int cid,
            ChannelNameSearchResultProcessor resultProcessor, long initialDelay);

    /**
     * Cancels the search for a channel. The cancellation may be processed
     * asynchronously. This means that the
     * {@link ChannelNameSearchResultProcessor} that has been registered for the
     * specified channel might still be called after calling this method. If no
     * search request has been queued for the specified channel, the search has
     * already succeeded, or the search request has already been cancelled,
     * calling this method has no effect.
     * 
     * @param cid
     *            client-assigned numeric identified for the channel that was
     *            specified when queuing the search request (never
     *            <code>null</code>).
     */
    void cancelChannelName(int cid);

}
