/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.client.resolver.internal;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.charset.Charset;

import com.aquenos.epics.jackie.client.beacon.BeaconDetector;
import com.aquenos.epics.jackie.client.io.ChannelAccessServerConnection;
import com.aquenos.epics.jackie.common.exception.ErrorHandler;
import com.aquenos.epics.jackie.common.io.CommunicationController;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessConstants;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessMessageCodec;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessSearchClientMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessSearchTCPServerMessage;
import com.aquenos.epics.jackie.common.protocol.ChannelAccessVersion;

/**
 * <p>
 * Connection with a Channel Access name server. The connection is responsible
 * for sending and receiving messages.
 * </p>
 * 
 * <p>
 * This class is only intended for use by {@link TcpResolver}. It is closely
 * tied to that class and has only been separated into a class of its own to
 * keep {@link TcpResolver} smaller.
 * </p>
 * 
 * <p>
 * This class is not thread safe, but the instantiating {@link TcpResolver}
 * ensures that it is only used in a safe way.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public class TcpNameServerConnection extends ChannelAccessServerConnection {

    private ErrorHandler errorHandler;
    private boolean ready = false;
    private TcpResolver resolver;
    private InetSocketAddress serverAddress;

    /**
     * Creates a connection. The channel backing the connection is connected
     * automatically. However, the connection process may finish asynchronously.
     * 
     * @param resolver
     *            the TCP resolver owning this connection.
     * @param beaconDetector
     *            beacon detector or <code>null</code> if beacons should not be
     *            used to verify the connection health.
     * @param charset
     *            character set to use for serializing channel names.
     * @param communicationController
     *            communication controller with which the channel backing this
     *            connection is registered.
     * @param errorHandler
     *            error handler used for logging unexpected exceptions.
     * @param serverAddress
     *            IP address and port number of the name server.
     * @throws IOException
     *             if an error occurs while creating and configuring the socket
     *             channel.
     */
    public TcpNameServerConnection(TcpResolver resolver,
            BeaconDetector beaconDetector, Charset charset,
            CommunicationController communicationController,
            ErrorHandler errorHandler, InetSocketAddress serverAddress)
            throws IOException {
        // For name resolution, a maximum message size of 16384 bytes is by far
        // enough. If limit for channel names is much shorter anyway.
        // Channel Access version 4.4 is the oldest version that we support. The
        // server might be even older, but this is okay because it has no
        // consequence for the start of the connection and the connection is
        // going to be destroyed automatically, if the server version is older.
        super(beaconDetector, charset, communicationController, errorHandler,
                ChannelAccessMessageCodec.MIN_MAX_PAYLOAD_SIZE
                        + ChannelAccessConstants.STANDARD_HEADER_SIZE,
                ChannelAccessMessageCodec.MIN_MAX_PAYLOAD_SIZE
                        + ChannelAccessConstants.STANDARD_HEADER_SIZE,
                serverAddress, ChannelAccessVersion.V4_4);
        assert (resolver != null);
        this.errorHandler = errorHandler;
        this.resolver = resolver;
        this.serverAddress = serverAddress;
    }

    @Override
    protected void onDestroy() {
        // Call the implementation in the super-class so that it can release
        // its resources.
        super.onDestroy();
        // Notify the TCP resolver that the connection has been destroyed.
        resolver.serverDestroyed(serverAddress);
    }

    @Override
    protected void onMessageReceived(ChannelAccessMessage message) {
        switch (message.getCommand()) {
        case CA_PROTO_SEARCH:
            resolver.receivedSearchResponse(
                    (ChannelAccessSearchTCPServerMessage) message, getVersion());
            break;
        case CA_PROTO_VERSION:
            // Name resolution over TCP connections is only supported since
            // Channel Access version 4.12. If the server is too old, we do
            // not mark this connection as ready but do not destroy it either.
            // This will keep the TCP resolver from trying to connect again.
            // When the connection is interrupted, we actually want to
            // reconnect, because the protocol version on the server might have
            // changed.
            if (getVersion().includes(ChannelAccessVersion.V4_12)) {
                ready = true;
                resolver.serverAvailable(serverAddress);
            }
            break;
        case CA_PROTO_ECHO:
            // We ignore echo messages. They have already been handled.
            break;
        default:
            errorHandler.handleError(TcpNameServerConnection.class, null,
                    "Received an unexpected message: " + message);
        }
    }

    /**
     * Tells whether the connection is ready to process search requests. The
     * connection is ready when it is connected and the protocol version has
     * been received from the server.
     * 
     * @return <code>true</code> if the connection is ready to process search
     *         requests, <code>false</code> if it is not ready (for example
     *         because it has been destroyed).
     */
    public boolean isAvailable() {
        return ready && !isDestroyed();
    }

    @Override
    public void destroy() {
        // We only override this method to make it public.
        super.destroy();
    }

    @Override
    public boolean isDestroyed() {
        // We only override this method to make it public.
        return super.isDestroyed();
    }

    /**
     * Sends a search message to the server. This message may only be called
     * when the connection is ready ({@link #isAvailable()} returns
     * <code>true</code>). If the connection has been destroyed, this method
     * does nothing.
     * 
     * @param message
     *            search message to be sent.
     */
    public void sendSearchMessage(ChannelAccessSearchClientMessage message) {
        // We only use an assertion here because the calling code should check
        // this.
        assert (ready);
        // In theory, this method might throw. In reality, this will not happen
        // because the channel name will always be short enough (otherwise,
        // trying to send the message via UDP would already have failed).
        sendMessage(message);
    }

}
