/*
 * Copyright 2014-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import java.nio.ByteBuffer;

/**
 * <p>
 * Abstract base class for {@link ByteSink} implementations. This base class
 * implements most of the methods declared in the {@link ByteSink} interface.
 * However, the typed put operations in this class incur an extra copy operation
 * in order to convert the raw data-type to bytes.
 * </p>
 * <p>
 * If possible, implementations should override the typed put operations with a
 * more efficient implementation that can write the data type directly. Usually,
 * it is sufficient to override the operations that take an array as well as an
 * offset and length, because those methods a called by the convenience methods
 * that take a single value or an array without additional parameters. However,
 * derived classes are also free to override the other methods, if they can
 * provide a more efficient implementation for them.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public abstract class AbstractByteSink implements ByteSink {

    /**
     * Size of a byte in bytes.
     */
    protected final static int SIZE_OF_BYTE = Byte.SIZE / 8;

    /**
     * Size of a double in bytes.
     */
    protected final static int SIZE_OF_DOUBLE = Double.SIZE / 8;

    /**
     * Size of a float in bytes.
     */
    protected final static int SIZE_OF_FLOAT = Float.SIZE / 8;

    /**
     * Size of an int in bytes.
     */
    protected final static int SIZE_OF_INT = Integer.SIZE / 8;

    /**
     * Size of a long in bytes.
     */
    protected final static int SIZE_OF_LONG = Long.SIZE / 8;

    /**
     * Size of a short in bytes.
     */
    protected final static int SIZE_OF_SHORT = Short.SIZE / 8;

    /**
     * Default constructor. This constructor is supposed to be used by classes
     * that inherit from this class.
     */
    protected AbstractByteSink() {
    }

    private void checkSourceArrayLength(int arrayLength, int offset, int length) {
        if (offset < 0) {
            throw new IllegalArgumentException("Offset must not be negative.");
        }
        if (length < 0) {
            throw new IllegalArgumentException("Length must not be negative.");
        }
        if ((long) offset + (long) length > (long) arrayLength) {
            throw new IndexOutOfBoundsException();
        }
    }

    @Override
    public ByteSink putByte(byte b) {
        return putByteArray(new byte[] { b });
    }

    @Override
    public ByteSink putByteArray(byte[] source) {
        return putByteArray(source, 0, source.length);
    }

    @Override
    public ByteSink putDouble(double value) {
        return putDoubleArray(new double[] { value });
    }

    @Override
    public ByteSink putDoubleArray(double[] source) {
        return putDoubleArray(source, 0, source.length);
    }

    @Override
    public ByteSink putDoubleArray(double[] source, int offset, int length) {
        checkSourceArrayLength(source.length, offset, length);
        byte[] byteArray = new byte[SIZE_OF_DOUBLE * length];
        ByteBuffer buffer = ByteBuffer.wrap(byteArray);
        buffer.asDoubleBuffer().put(source, offset, length);
        return putByteArray(byteArray);
    }

    @Override
    public ByteSink putFloat(float value) {
        return putFloatArray(new float[] { value });
    }

    @Override
    public ByteSink putFloatArray(float[] source) {
        return putFloatArray(source, 0, source.length);
    }

    @Override
    public ByteSink putFloatArray(float[] source, int offset, int length) {
        checkSourceArrayLength(source.length, offset, length);
        byte[] byteArray = new byte[SIZE_OF_FLOAT * length];
        ByteBuffer buffer = ByteBuffer.wrap(byteArray);
        buffer.asFloatBuffer().put(source, offset, length);
        return putByteArray(byteArray);
    }

    @Override
    public ByteSink putInt(int value) {
        return putIntArray(new int[] { value });
    }

    @Override
    public ByteSink putIntArray(int[] source) {
        return putIntArray(source, 0, source.length);
    }

    @Override
    public ByteSink putIntArray(int[] source, int offset, int length) {
        checkSourceArrayLength(source.length, offset, length);
        byte[] byteArray = new byte[SIZE_OF_INT * length];
        ByteBuffer buffer = ByteBuffer.wrap(byteArray);
        buffer.asIntBuffer().put(source, offset, length);
        return putByteArray(byteArray);
    }

    @Override
    public ByteSink putShort(short value) {
        return putShortArray(new short[] { value });
    }

    @Override
    public ByteSink putShortArray(short[] source) {
        return putShortArray(source, 0, source.length);
    }

    @Override
    public ByteSink putShortArray(short[] source, int offset, int length) {
        checkSourceArrayLength(source.length, offset, length);
        byte[] byteArray = new byte[SIZE_OF_SHORT * length];
        ByteBuffer buffer = ByteBuffer.wrap(byteArray);
        buffer.asShortBuffer().put(source, offset, length);
        return putByteArray(byteArray);
    }

    @Override
    public ByteSink putLong(long value) {
        return putLongArray(new long[] { value });
    }

    @Override
    public ByteSink putLongArray(long[] source) {
        return putLongArray(source, 0, source.length);
    }

    @Override
    public ByteSink putLongArray(long[] source, int offset, int length) {
        checkSourceArrayLength(source.length, offset, length);
        byte[] byteArray = new byte[SIZE_OF_LONG * length];
        ByteBuffer buffer = ByteBuffer.wrap(byteArray);
        buffer.asLongBuffer().put(source, offset, length);
        return putByteArray(byteArray);
    }

}
