/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.io;

import java.io.IOException;
import java.net.SocketException;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.channels.CancelledKeyException;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.concurrent.atomic.AtomicBoolean;

import com.aquenos.epics.jackie.common.exception.ErrorHandler;

/**
 * <p>
 * Abstract base class for handling a connection based on a
 * {@link SocketChannel}.
 * </p>
 * 
 * <p>
 * This class implements the low-level details of dealing with a non-blocking
 * socket-channel. Classes inherting from this class can simply write to a
 * {@link ByteSink} and read from a {@link ByteSource}. This class takes care of
 * registering the right operations with the channel's {@link Selector} and
 * managing internal buffers.
 * </p>
 * 
 * <p>
 * This class is not safe for use by multiple threads. All of its methods
 * (except for the {@link #destroy()} method) may only be called from the
 * communication controller's thread. This can easily be achieved by
 * implementing all logic in the event methods (<code>on...</code>).
 * </p>
 * 
 * <p>
 * This class is highly configurable. The size of an individual
 * {@link ByteBuffer}, the number of unused byte buffers that are kept on a free
 * list, and the maximum number of bytes that are read contiguously can be
 * configured at construction time.
 * </p>
 * 
 * @author Sebastian Marsching
 */
public abstract class AbstractSocketChannelConnection {

    private int bufferSize;
    private SocketChannel channel;
    private SelectionKey channelSelectionKey;
    private CommunicationController communicationController;
    private AtomicBoolean destroyed = new AtomicBoolean(false);
    private ErrorHandler errorHandler;
    private LinkedList<ByteBuffer> freeBuffers = new LinkedList<ByteBuffer>();
    private int maximumNumberOfFreeBuffers;
    private ByteBuffer receiveBuffer;
    private long receiveMaxContiguousBytes = 0L;
    private long receiveBytesNeeded = 0L;
    private ByteBufferByteSource receiveSource = new ByteBufferByteSource() {

        @Override
        protected void release(ByteBuffer buffer) {
            // This is not the best point to test which thread we are running
            // in, but it should trigger an assertion error at some point when
            // the wrong thread reads from the byte source.
            assert (communicationController.inCommunicationThread());
            releaseBuffer(buffer);
        }

    };
    private LinkedList<ByteBuffer> sendQueue = new LinkedList<ByteBuffer>();
    private long sendQueueBytes = 0L;
    private ByteBufferByteSink sendSink = new ByteBufferByteSink() {

        @Override
        protected void dataWritten() {
            // This is not the best point to test which thread we are running
            // in, but it should trigger an assertion error at some point when
            // the wrong thread writes to the byte sink.
            assert (communicationController.inCommunicationThread());
            // Data is available now, so we want to write it, when possible.
            enableWriteEvent();
        }

        @Override
        protected ByteBuffer requestAdditionalBuffer(int requiredSpaceInBytes) {
            // This is not the best point to test which thread we are running
            // in, but it should trigger an assertion error at some point when
            // the wrong thread writes to the byte sink.
            assert (communicationController.inCommunicationThread());
            return acquireBuffer();
        }

    };

    /**
     * Creates a connection based on a socket channel. The <code>channel</code>
     * has to be configured for non-blocking operation and must be connected or
     * a connection must be pending.
     * 
     * @param channel
     *            socket channel that represents the connection (never
     *            <code>null</code>).
     * @param communicationController
     *            communication controller used by this connection (never
     *            <code>null</code>).
     * @param errorHandler
     *            error handler that is used to log unexpected I/O errors (never
     *            <code>null</code>). Only errors that are typically not caused
     *            by an interrupted connection are considered unexpected.
     * @param bufferSize
     *            size of each byte buffer that is allocated by this class. Must
     *            be greater than zero and should (but does not have to be) a
     *            power of two. Larger buffers typically improve the performance
     *            but can have a large memory footprint if they are not filled
     *            completely. However, this class avoids partially filled
     *            buffers by reusing them whenever possible. Therefore, the
     *            impact on memory usage should be negligible.
     * @param maximumNumberOfFreeBuffers
     *            maximum number of unused buffers to keep on the free list.
     *            Must be greater than or equal to zero. This class uses direct
     *            buffers which typically incur significant allocation cost.
     *            Therefore, allowing for a larger number of unused buffers to
     *            be kept can improve the performance but might increase the
     *            memory footprint.
     * @param maximumNumberOfContiguousBytesRead
     *            maximum number of bytes that are read in a single operation.
     *            If zero or negative, the number is not limited. This parameter
     *            can be used to avoid staying in the receive loop forever when
     *            data arrives faster than it can be processed. When a limit is
     *            set and the receive loop detects that the last read operation
     *            hit or surpassed this limit, it will quit the loop, thus
     *            returning control back to the communication controller. This
     *            way, the communication controller is able to process other
     *            pending actions, including, but not limited to, send
     *            operations for this object.
     * 
     * @throws IllegalArgumentException
     *             if one of the parameters passed to this method is invalid.
     *             Invalid parameters are a <code>channel</code> that is not
     *             configured for non-blocking operation or that is not
     *             connected and no pending connection attempt, a
     *             <code>bufferSize</code> less than one, and a
     *             <code>maximumNumberOfFreeBuffers</code> less than zero.
     */
    public AbstractSocketChannelConnection(SocketChannel channel,
            CommunicationController communicationController,
            ErrorHandler errorHandler, int bufferSize,
            int maximumNumberOfFreeBuffers,
            long maximumNumberOfContiguousBytesRead) {
        this.bufferSize = bufferSize;
        if (this.bufferSize < 1) {
            throw new IllegalArgumentException(
                    "Buffer size must be greater than or equal to one.");
        }
        this.channel = channel;
        if (this.channel.isBlocking()) {
            throw new IllegalArgumentException(
                    "The socket channel must be configured for non-blocking I/O.");
        }
        if (!this.channel.isConnected() && !this.channel.isConnectionPending()) {
            throw new IllegalArgumentException(
                    "The channel must be connect or a connection must be pending.");
        }
        this.communicationController = communicationController;
        this.errorHandler = errorHandler;
        if (this.errorHandler == null) {
            throw new NullPointerException();
        }
        this.maximumNumberOfFreeBuffers = maximumNumberOfFreeBuffers;
        if (this.maximumNumberOfFreeBuffers < 0) {
            throw new IllegalArgumentException(
                    "Maximum number of free buffers must be greater than or equal to zero.");
        }
        this.receiveMaxContiguousBytes = maximumNumberOfContiguousBytesRead;
        // Register the channel with the communication controller.
        try {
            this.channelSelectionKey = this.communicationController
                    .registerChannel(new ChannelProcessor() {
                        @Override
                        public void processIO() {
                            AbstractSocketChannelConnection.this.processIO();
                        }
                    }, this.channel);
        } catch (ClosedChannelException e) {
            // This only happens when this object has already been destroyed.
            throw new IllegalArgumentException(
                    "The channel must not be closed.", e);
        }
        // If we are in the communication thread, we can finish the
        // initialization directly. Otherwise, we have to do this later, when we
        // are in the communication thread.
        if (this.communicationController.inCommunicationThread()) {
            finishInitialization();
        } else {
            this.communicationController.registerTimer(new TimerProcessor() {
                @Override
                public void processTimer() {
                    finishInitialization();
                }
            }, 0L);
        }
    }

    private void finishInitialization() {
        // Here we do all initialization that has to be done in the
        // communication thread in order to avoid a race condition.
        if (channel.isConnected()) {
            // If the channel is already connected, we trigger onConnect. We can
            // do this safely because processIO cannot have been called yet
            // (there interestOps are still empty).
            onConnect();
        }
        // We always want to be notified when we can receive data.
        channelSelectionKey.interestOps(SelectionKey.OP_READ);
        // If the channel is not connected yet, we want to be notified when the
        // connection has been established.
        if (channel.isConnectionPending()) {
            enableConnectEvent();
        }
    }

    private ByteBuffer acquireBuffer() {
        ByteBuffer buffer = freeBuffers.poll();
        if (buffer != null) {
            return buffer;
        } else {
            return ByteBuffer.allocateDirect(bufferSize);
        }
    }

    private void releaseBuffer(ByteBuffer buffer) {
        if (freeBuffers.size() < maximumNumberOfFreeBuffers) {
            buffer.clear();
            freeBuffers.push(buffer);
        }
    }

    private void enableConnectEvent() {
        channelSelectionKey.interestOps(channelSelectionKey.interestOps()
                | SelectionKey.OP_CONNECT);
    }

    private void disableConnectEvent() {
        channelSelectionKey.interestOps(channelSelectionKey.interestOps()
                & ~SelectionKey.OP_CONNECT);
    }

    private void enableWriteEvent() {
        channelSelectionKey.interestOps(channelSelectionKey.interestOps()
                | SelectionKey.OP_WRITE);
    }

    private void disableWriteEvent() {
        channelSelectionKey.interestOps(channelSelectionKey.interestOps()
                & ~SelectionKey.OP_WRITE);
    }

    private void handleIOException(IOException e) {
        if (e instanceof SocketException) {
            // A SocketException has to be expected. It is most likely caused by
            // the peer closing the connection or a problem while connecting. We
            // do not want to log this kind of exception.
        } else {
            // If the IOException is of an unexpected type, we log it. We use
            // the class's actual type because this is a base class that is not
            // useful on its own.
            errorHandler.handleError(this.getClass(), e,
                    "Unexpected I/O exception for socket channel.");
        }
        // We always destroy the connection because for a SocketChannel, an
        // IOException typically is non-recoverable.
        destroy();
    }

    /**
     * <p>
     * Tries to read data from the channel. This method reads, until there is no
     * more data available (the read returns no data), there is an
     * {@link IOException}, or the maximum number of contiguous bytes to be
     * received is reached. When data is received. the {@link #onReceive()}
     * method is called. Before returning, the
     * {@link #onReceiveFinished(boolean)} is called.
     * </p>
     * 
     * <p>
     * This method catches all {@link IOException}s. If such an exception is
     * caught, the connection is destroyed by calling {@link #destroy()}. If the
     * connection is pending (a connect operation has been started but has not
     * finished yet), this method does not try to read but returns immediately.
     * {@link #onReceiveFinished(boolean)} is not called.
     * </p>
     * 
     * <p>
     * Typically, this method does not have to be called explicitly. It will
     * automatically be triggered, when data becomes available for reading.
     * However, derived classes can call it explicitly if they want to run this
     * method when it would normally not be run (e.g. when no data is available
     * or when other operations like sending data would usually be run first).
     * </p>
     * 
     * <p>
     * This method may only be called from the communication controller's
     * thread.
     * </p>
     */
    protected void tryReceive() {
        assert (communicationController.inCommunicationThread());
        // We check for connection pending, not connected. This is by intention
        // because we want to continue if the channel has been connected but is
        // not connected any longer. This will cause an exception which will
        // cause this object to be destroyed.
        if (channel.isConnectionPending()) {
            // The channel is not connected yet, thus we cannot read any data.
            return;
        }
        long totalBytesRead = 0L;
        boolean dataReceived;
        do {
            dataReceived = false;
            if (receiveBuffer == null) {
                receiveBuffer = acquireBuffer();
            }
            int bytesRead;
            try {
                bytesRead = channel.read(receiveBuffer);
            } catch (IOException e) {
                handleIOException(e);
                return;
            }
            // If we hit end of stream, the connection got closed so we destroy
            // it.
            if (bytesRead == -1) {
                destroy();
                return;
            }
            if (bytesRead != 0) {
                dataReceived = true;
                // When we received enough data, we reset the number of bytes
                // needed so that the onReceive method gets called.
                if (receiveBytesNeeded < bytesRead) {
                    receiveBytesNeeded = 0L;
                } else {
                    receiveBytesNeeded -= bytesRead;
                }
            }
            if (Long.MAX_VALUE - totalBytesRead < bytesRead) {
                totalBytesRead = Long.MAX_VALUE;
            } else {
                totalBytesRead += bytesRead;
            }
            // If we received data and we received enough data since the last
            // call to onReceive, we want to call onReceive again.
            if (dataReceived && receiveBytesNeeded == 0L) {
                // We have to remember whether there is space left in the
                // buffer because we want to reuse it in this case.
                boolean reuseBuffer = receiveBuffer.hasRemaining();
                receiveBuffer.flip();
                receiveSource.appendBuffer(receiveBuffer);
                receiveBuffer = null;
                // If onReceive returns a positive number, this is the number of
                // bytes that we should receive (in addition to the number of
                // bytes that we already received)
                receiveBytesNeeded = onReceive();
                if (receiveBytesNeeded < 0L) {
                    receiveBytesNeeded = 0L;
                }
                // If there was space left in the buffer, we want to reuse the
                // buffer. This way, we can avoid to accumulate partially filled
                // buffers. This saves memory and decreases the chances of
                // unaligned read operations.
                if (reuseBuffer) {
                    receiveBuffer = receiveSource.removeLastBuffer();
                    // If all data has been used, the receiveBuffer is null.
                    // This is okay because it means that it has already been
                    // released.
                    if (receiveBuffer != null) {
                        if (receiveBuffer.position() != 0) {
                            // Some data has already been read. We have to
                            // compact the buffer, so that it does not get read
                            // again.
                            receiveBuffer.compact();
                        } else {
                            // We have to set the position so that the data that
                            // has not been read yet does not get overwritten.
                            receiveBuffer.position(receiveBuffer.limit());
                            receiveBuffer.limit(receiveBuffer.capacity());
                        }
                    }
                }
            }
        } while (dataReceived
                && (receiveMaxContiguousBytes <= 0L || receiveMaxContiguousBytes > totalBytesRead));
        // Notify a derived class that the receive operation finished.
        onReceiveFinished(dataReceived);
    }

    /**
     * <p>
     * Tries to write data to the channel. This method writes, until there is no
     * more data available (the send sink is empty), there is an
     * {@link IOException}, or the channel does not accept more data. Before
     * returning, the {@link #onSend()} method is called if any data has been
     * written.
     * </p>
     * 
     * <p>
     * This method catches all {@link IOException}s. If such an exception is
     * caught, the connection is destroyed by calling {@link #destroy()}. If the
     * connection is pending (a connect operation has been started but has not
     * finished yet), this method does not try to write but returns immediately.
     * </p>
     * 
     * <p>
     * Typically, this method does not have to be called explicitly. It will
     * automatically be triggered, when there is data in the send sink and the
     * channel is ready for writing. However, derived classes can call it
     * explicitly if they want to run this method when it would normally not be
     * run (e.g. or when other operations like receiving data would usually be
     * run first).
     * </p>
     * 
     * <p>
     * This method may only be called from the communication controller's
     * thread.
     * </p>
     */
    protected void trySend() {
        assert (communicationController.inCommunicationThread());
        // We check for connection pending, not connected. This is by intention
        // because we want to continue if the channel has been connected but is
        // not connected any longer. This will cause an exception which will
        // cause this object to be destroyed.
        if (channel.isConnectionPending()) {
            // The channel is not connected yet, thus we cannot write any data.
            return;
        }
        // Get data from the byte sink.
        long writtenDataBytes = sendSink.getWrittenDataSizeInBytes();
        sendQueue.addAll(Arrays.asList(sendSink.getWrittenData()));
        if (Long.MAX_VALUE - sendQueueBytes < writtenDataBytes) {
            sendQueueBytes = Long.MAX_VALUE;
        } else {
            sendQueueBytes += writtenDataBytes;
        }
        if (sendQueue.isEmpty()) {
            disableWriteEvent();
            return;
        }
        boolean dataWritten = false;
        ByteBuffer sendBuffer = sendQueue.poll();
        try {
            int bytesSent;
            do {
                bytesSent = channel.write(sendBuffer);
                if (sendQueueBytes != Long.MAX_VALUE) {
                    sendQueueBytes -= bytesSent;
                }
                if (bytesSent != 0) {
                    dataWritten = true;
                }
                if (!sendBuffer.hasRemaining()) {
                    releaseBuffer(sendBuffer);
                    sendBuffer = sendQueue.poll();
                }
            } while (bytesSent != 0 && sendBuffer != null);
        } catch (IOException e) {
            handleIOException(e);
            return;
        }
        // If we quit the loop because we could not send any more data (and not
        // because we sent all send buffers), we have to put the last buffer
        // back into the queue so that it is sent the next time. We also enable
        // the write event, because it might not be enabled yet. If we do not
        // have any more data to send, we can disable the write event because
        // there is no sense in calling this method if there is no data.
        if (sendBuffer != null) {
            sendQueue.push(sendBuffer);
            enableWriteEvent();
        } else {

            disableWriteEvent();
        }
        // If we did not write anything from the last buffer in the queue, we
        // want to add it to the byte sink again. This way, it can be filled
        // completely instead of having a partially filled buffer lingering
        // around.
        ByteBuffer lastBuffer = sendQueue.peekLast();
        if (lastBuffer != null && lastBuffer.hasRemaining()) {
            sendQueue.removeLast();
            if (lastBuffer.position() == 0) {
                lastBuffer.position(lastBuffer.limit());
                lastBuffer.limit(lastBuffer.capacity());
            } else {
                lastBuffer.compact();
            }
            sendSink.addBuffer(lastBuffer);
        }
        // If we sent any data, we want to trigger the onSend method.
        if (dataWritten) {
            onSend();
        }
    }

    /**
     * Tells whether this connection has been destroyed. A connection is
     * destroyed when its {@link #destroy()} method is called. The destroy
     * method is called when an I/O exception occurs, but it might also be
     * called explicitly by derived classes.
     * 
     * @return <code>true</code> if this connection has been destroyed,
     *         <code>false</code> if it has not been destroyed yet.
     */
    protected boolean isDestroyed() {
        return destroyed.get();
    }

    /**
     * Returns the byte source that contains the data received by this
     * connection. This byte source may only be used from the communication
     * controller's thread. The byte source returned implements the
     * {@link ByteSource#remaining() remaining()} method accurately. This means
     * that reading more bytes than indicated by this method will always result
     * in a {@link BufferUnderflowException}. If the method indicates that there
     * are {@link Integer#MAX_VALUE} bytes remaining, the actual number of bytes
     * remaining might be larger.
     * 
     * @return byte source containing received data (never <code>null</code>).
     */
    protected ByteSource getReceiveSource() {
        return receiveSource;
    }

    /**
     * Returns the byte sink that can be used to send data over this connection.
     * This byte sink may only be used from the communication controller's
     * thread. Data written to this byte sink is sent automatically.
     * 
     * @return byte sink used for sending data (never <code>null</code>).
     */
    protected ByteSink getSendSink() {
        return sendSink;
    }

    /**
     * Returns the number of bytes that have been written to the byte sink but
     * have not been sent yet. If the number is greater than
     * {@link Long#MAX_VALUE}, this value is returned. The number returned can
     * be used to determine whether more data should be added to the queue or
     * whether this should be postponed until some of the data in the queue has
     * been sent. This method may only be called from the communication
     * controller's thread.
     * 
     * @return number of bytes that have been written to the byte sink but have
     *         not been written to the channel yet.
     */
    protected long getSendQueueSizeInBytes() {
        assert (communicationController.inCommunicationThread());
        if (sendQueueBytes == Long.MAX_VALUE) {
            // The length of the send queue might not be accurate because we
            // stopped keeping track when we hit the limit. Therefore, we have
            // to calculate the length.
            sendQueueBytes = 0L;
            for (ByteBuffer buffer : sendQueue) {
                int remaining = buffer.remaining();
                if (Long.MAX_VALUE - sendQueueBytes < remaining) {
                    sendQueueBytes = Long.MAX_VALUE;
                    return Long.MAX_VALUE;
                } else {
                    sendQueueBytes += remaining;
                }
            }
        }
        long sendSinkBytes = sendSink.getWrittenDataSizeInBytes();
        if (Long.MAX_VALUE - sendQueueBytes < sendSinkBytes) {
            return Long.MAX_VALUE;
        } else {
            return sendQueueBytes + sendSinkBytes;
        }
    }

    /**
     * Called when the socket channel is connected. This method is called
     * exactly once, when the socket changes from the connection pending to the
     * connected state. If the channel is already connected when this object is
     * created, this method is still called once. Therefore, child classes can
     * always use this method to implement logic that should be run once, after
     * a connection has been established. This method is guaranteed to be called
     * in the communication controller's thread.
     */
    protected void onConnect() {
        // Empty connect event handler. Can be overridden to trigger an action
        // when the connection is established.
    }

    /**
     * Called when this connection has been destroyed. A connection can be
     * destroyed by two events: If a read or write operation on the channel
     * fails, or if the {@link #destroy()} method is called explicitly. This
     * method will never be called more than once. This method is called in the
     * thread which calls the {@link #destroy()} method.
     */
    protected void onDestroy() {
        // Empty destroy event handler. Can be overridden to trigger an action
        // when the connection is destroyed.
    }

    /**
     * Called when data is received. The implementation provided by the base
     * class simply discards all data received. Derived classes should override
     * this method in order to process the received data. All received data has
     * been added to the byte source, which can be retrieved by calling
     * {@link #getReceiveSource()}. If this method returns a number greater than
     * zero, it will only be called again when at least this number of bytes
     * have been received (in addition to the bytes that already have been
     * received and are available from the byte source). This method is
     * guaranteed to be called in the communication controller's thread.
     * 
     * @return number of additional bytes needed to do further processing or
     *         zero if this method should be called again as soon as any number
     *         of additional bytes arrive.
     */
    protected long onReceive() {
        // We simply drain all received data. Should be overridden in order to
        // actually process received data.
        int remaining;
        while ((remaining = receiveSource.remaining()) != 0) {
            receiveSource.skip(remaining);
        }
        return 0L;
    }

    /**
     * Called when a data receive operation is finished. A receive operation
     * might finish because of two reasons. Either there is no more data that
     * can be read from the channel or the limit for maximum number of bytes
     * that may be read contiguously has been reached. Derived classes can
     * implement this method in order to be informed when a receive operation is
     * interrupted because there is too much data or when a receive operation
     * stops because there is no more data available. This method is guaranteed
     * to be called in the communication controller's thread.
     * 
     * @param moreDataAvailable
     *            <code>true</code> if more data is available from the channel
     *            but the read operation has been interrupted because the limit
     *            for the maximum number of contiguous bytes read has been
     *            reached. <code>false</code> if currently there is no more data
     *            available from the channel.
     */
    protected void onReceiveFinished(boolean moreDataAvailable) {
        // Can be overridden in derived classes.
    }

    /**
     * Called whenever data has been sent. This method is called when at least
     * one byte of data has been written to this connection's channel and now no
     * further data is written (either because the send queue is empty or
     * because the channel does not accept more data at the moment). This method
     * can be overridden by derived classes in order to perform an action when
     * data has been sent successfully. In general, this can be treated as a
     * progress indicator. The number of bytes remaining to be sent can be
     * retrieved with {@link #getSendQueueSizeInBytes()}. This method is
     * guaranteed to be called in the communication controller's thread.
     */
    protected void onSend() {
        // Empty send event handler. Can be overridden to trigger an action when
        // data is actually sent.
    }

    private void processIO() {
        try {
            // For some reason, OP_CONNECT is not set in readyOps() on JRE 6
            // when finishConnect() should be called, however the select
            // operation still succeeds. This bug is triggered when the
            // selector's wakeup method is called before adding OP_CONNECT to
            // the interestOps. It does not matter whether wakeup is called
            // before or after registering the channel with the selector, just
            // whether before or after registering OP_CONNECT with the
            // interestOps(). As wakeup is used quite frequently when dealing
            // with the communication controller outside the communication
            // thread (e.g. when registering a channel or a timer), we have to
            // work around this bug.
            // Luckily, there is a simple solution: We do not check whether
            // OP_CONNECT is set but simply try finishConnect() when this method
            // gets called and the channel is not connected. This does not have
            // a negative effect, because it will only happen until the channel
            // is connected and finishConnect() is non-blocking.
            if ((channelSelectionKey.interestOps() & SelectionKey.OP_CONNECT) != 0) {
                try {
                    if (!channel.isConnected() && channel.finishConnect()) {
                        onConnect();
                    }
                } catch (IOException e) {
                    handleIOException(e);
                    return;
                }
                if (channel.isConnected()) {
                    disableConnectEvent();
                }
            }
            // We try to send data before receiving it. Received data might
            // cause more data to be sent and thus the send queue would just
            // grow more than it has to if we did it the other way round.
            if ((channelSelectionKey.readyOps() & SelectionKey.OP_WRITE) != 0) {
                trySend();
            }
            if ((channelSelectionKey.readyOps() & SelectionKey.OP_READ) != 0) {
                tryReceive();
            }
        } catch (CancelledKeyException e) {
            // A CancelledKeyException can occur because the destroy method that
            // closes the channel and cancels the selection key may be called
            // from any thread. We do not consider such an exception an error.
            // It just means that processing I/O does not make sense any longer,
            // so we simply return.
            return;
        }
    }

    /**
     * Destroys this connection. Destroying the connection closes the
     * connection's channel and triggers the {@link #onDestroy()} method. This
     * method may be called from any thread. If the connection has already been
     * destroyed, calling this method has no effect. The <code>onDestroy</code>
     * method is called before this method returns.
     */
    protected void destroy() {
        // We only want to destroy this object once (in particular, we do not
        // want to call the onDestroy method more than once). We use an atomic
        // boolean to check, whether this object has already been destroyed.
        if (destroyed.get()) {
            return;
        }
        if (!destroyed.compareAndSet(false, true)) {
            return;
        }
        channelSelectionKey.cancel();
        try {
            channel.close();
        } catch (IOException e) {
            // We ignore any exception that occurs while trying to close the
            // channel.
        }
        onDestroy();
    }

}
