/*
 * Copyright 2014-2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

/**
 * Command code that can be specified in the header of a Channel Access message.
 * The command defines the type of the message and thus the meaning of all other
 * header fields.
 * 
 * @author Sebastian Marsching
 */
public enum ChannelAccessCommand {

    /**
     * Message for communicating the protocol version.. Messages of this type
     * are used to exchange the supported minor version number and (depending on
     * the context) sequence numbers and priorities.
     * 
     * @see ChannelAccessVersionMessage
     * @see ChannelAccessVersionTCPClientMessage
     * @see ChannelAccessVersionTCPServerMessage
     * @see ChannelAccessVersionUDPMessage
     */
    CA_PROTO_VERSION(0),

    /**
     * Message for creating an event subscription. Messages of this type are
     * also used to send subscription events or to confirm the cancellation of a
     * subscription.
     * 
     * @see ChannelAccessSubscriptionMessage
     * @see ChannelAccessSubscriptionClientMessage
     * @see ChannelAccessSubscriptionServerMessage
     * @see ChannelAccessCancelSubscriptionServerMessage
     */
    CA_PROTO_EVENT_ADD(1),

    /**
     * Message for canceling an event subscription.
     * 
     * @see ChannelAccessCancelSubscriptionClientMessage
     */
    CA_PROTO_EVENT_CANCEL(2),

    /**
     * Legacy message for reading from a channel. New implementations should
     * prefer the {@link #CA_PROTO_READ_NOTIFY} message.
     * 
     * @see ChannelAccessReadMessage
     * @see ChannelAccessReadClientMessage
     * @see ChannelAccessReadServerMessage
     */
    CA_PROTO_READ(3),

    /**
     * Message for writing to a channel. The client sends this message to the
     * server and does not expect the response. The server might (but does not
     * have to) reply with a {@link #CA_PROTO_ERROR} message if the write fails.
     * 
     * @see ChannelAccessWriteMessage
     */
    CA_PROTO_WRITE(4),

    /**
     * This message is not used by any supported version of the protocol.
     */
    CA_PROTO_SNAPSHOT(5),

    /**
     * Message for finding a channel. This message is used to look for a channel
     * as well as to respond to a channel find request.
     * 
     * @see ChannelAccessSearchMessage
     * @see ChannelAccessSearchClientMessage
     * @see ChannelAccessSearchTCPServerMessage
     * @see ChannelAccessSearchUDPServerMessage
     */
    CA_PROTO_SEARCH(6),

    /**
     * This message is not used by any supported version of the protocol.
     */
    CA_PROTO_BUILD(7),

    /**
     * Message for stopping the transmission of events. This message is sent by
     * a client if it wants to temporarily stop the flow of events for all event
     * subscriptions. The flow can later be resumed by sending
     * {@link #CA_PROTO_EVENTS_ON}.
     * 
     * @see ChannelAccessEventsOffMessage
     */
    CA_PROTO_EVENTS_OFF(8),

    /**
     * Message for resuming the transmission of events. This message is sent by
     * a client when it wants to resume the flow of event subscriptions after
     * sending {@link #CA_PROTO_EVENTS_OFF}.
     * 
     * @see ChannelAccessEventsOnMessage
     */
    CA_PROTO_EVENTS_ON(9),

    /**
     * Message for canceling outstanding read request. This message is sent by
     * the client to signal that it does not want to receive answers to read
     * requests it sent but have not been answered yet. It is also abused by
     * some clients as a replacement for the {@link #CA_PROTO_ECHO} message,
     * when using a protocol version that does not support the echo message yet.
     * 
     * @see ChannelAccessReadSyncMessage
     */
    CA_PROTO_READ_SYNC(10),

    /**
     * Message for signaling an error condition. This message is sent in reply
     * to a message that caused an error.
     * 
     * @see ChannelAccessErrorMessage
     */
    CA_PROTO_ERROR(11),

    /**
     * Message for closing a channel. This message is sent from the client to
     * the server in order to close a channel that has been opened using the
     * {@link #CA_PROTO_CREATE_CHAN} message.
     * 
     * @see ChannelAccessDisconnectChannelMessage
     */
    CA_PROTO_CLEAR_CHANNEL(12),

    /**
     * Message sent by a server to announce that it is running. This message is
     * periodically sent by servers in order to announce that they are running.
     * It can be used to detect new servers that have just been started.
     * 
     * @see ChannelAccessBeaconMessage
     */
    CA_PROTO_RSRV_IS_UP(13),

    /**
     * Message signaling that a channel could not be found. This message may be
     * sent in reply to a {@link #CA_PROTO_SEARCH} message, but only if the
     * client explicitly requested to be informed, if the server could not find
     * the channel. Usually, requests for channels that are not known by the
     * server are simply discarded.
     * 
     * @see ChannelAccessNotFoundMessage
     */
    CA_PROTO_NOT_FOUND(14),

    /**
     * Message for reading from a channel. This message is sent to read a
     * channel's current value. It is also sent in reply to such a read request.
     * If an event subscription is established for a channel, a message of this
     * type is sent for each event.
     * 
     * @see ChannelAccessReadNotifyMessage
     * @see ChannelAccessReadNotifyClientMessage
     * @see ChannelAccessReadNotifyServerMessage
     */
    CA_PROTO_READ_NOTIFY(15),

    /**
     * This message is not used by any supported version of the protocol.
     * 
     * @see ChannelAccessReadBuildMessage
     */
    CA_PROTO_READ_BUILD(16),

    /**
     * Message for confirming a registration with the Channel Access repeater.
     * It is sent by a CA repeater in order to acknowledge a registration with
     * that repeater.
     * 
     * @see #REPEATER_REGISTER
     * @see ChannelAccessRepeaterConfirmMessage
     */
    REPEATER_CONFIRM(17),

    /**
     * Message for opening a channel. This message is sent by the client in
     * order to open a channel. If the connection is established, the server
     * replies with a message of the same type (except for very old protocol
     * versions). If the connection cannot be established, the server replies
     * with a {@link #CA_PROTO_CREATE_CH_FAIL} or {@link #CA_PROTO_ERROR}
     * message.
     * 
     * @see ChannelAccessConnectChannelMessage
     * @see ChannelAccessConnectChannelClientMessage
     * @see ChannelAccessConnectChannelServerMessage
     */
    CA_PROTO_CREATE_CHAN(18),

    /**
     * Message for writing to a channel. This message is sent by the client in
     * order to write a value to a channel. The server replies with a message of
     * the same type when the write operation has finished. If the write
     * operation fails, the server might also send a {@link #CA_PROTO_ERROR}
     * message, depending on the cause of the error.
     * 
     * @see ChannelAccessWriteNotifyMessage
     * @see ChannelAccessWriteNotifyClientMessage
     * @see ChannelAccessWriteNotifyServerMessage
     */
    CA_PROTO_WRITE_NOTIFY(19),

    /**
     * Message for specifying the user name. This message is sent by the client
     * in order to communicate the user name. This information is used by some
     * servers as a very weak form of access control.
     * 
     * @see #CA_PROTO_HOST_NAME
     * @see #CA_PROTO_ACCESS_RIGHTS
     * @see ChannelAccessUserNameMessage
     */
    CA_PROTO_CLIENT_NAME(20),

    /**
     * Message for specifying the host name. This message is sent by the client
     * in order to communicate the host name. This information is used by some
     * servers as a very weak form of access control.
     * 
     * @see #CA_PROTO_CLIENT_NAME
     * @see #CA_PROTO_ACCESS_RIGHTS
     * @see ChannelAccessHostNameMessage
     */
    CA_PROTO_HOST_NAME(21),

    /**
     * Message for communicating the access rights for a channel. This message
     * is sent by the server in order to inform a client about the access rights
     * of a channel. The access rights might depend on the client's address as
     * well as the specified user name and host name.
     * 
     * @see #CA_PROTO_CLIENT_NAME
     * @see #CA_PROTO_HOST_NAME
     * @see ChannelAccessAccessRightsMessage
     */
    CA_PROTO_ACCESS_RIGHTS(22),

    /**
     * Message for testing the connection to a server. This message is sent by
     * the client in order to test the connection to a server. If the connection
     * works correctly, the server will respond immediately with a message of
     * the same type, which has the same headers and payload.
     * 
     * @see ChannelAccessEchoMessage
     */
    CA_PROTO_ECHO(23),

    /**
     * Legacy message for registering with a Channel Access repeater. Modern
     * clients simply send an empty UDP packet instead.
     * 
     * @see ChannelAccessRepeaterRegisterMessage
     */
    REPEATER_REGISTER(24),

    /**
     * This message is not used by any supported version of the protocol.
     */
    CA_PROTO_SIGNAL(25),

    /**
     * Message for indicating an error while trying to open a channel. This
     * message is sent in reply to a {@link #CA_PROTO_CREATE_CHAN} message, if
     * the channel cannot be opened due to a lack of resources on the server
     * side.
     * 
     * @see ChannelAccessConnectChannelFailedMessage
     */
    CA_PROTO_CREATE_CH_FAIL(26),

    /**
     * Message for indicating that a channel has been deleted from the server.
     * This message is sent by the server in order to indicate that a channel
     * has been deleted and the client should not try to access this channel any
     * longer.
     * 
     * @see ChannelAccessChannelDisconnectedByServerMessage
     */
    CA_PROTO_SERVER_DISCONN(27);

    private final static ChannelAccessCommand[] commandNumberToCommand = new ChannelAccessCommand[] {
            CA_PROTO_VERSION, CA_PROTO_EVENT_ADD, CA_PROTO_EVENT_CANCEL,
            CA_PROTO_READ, CA_PROTO_WRITE, CA_PROTO_SNAPSHOT, CA_PROTO_SEARCH,
            CA_PROTO_BUILD, CA_PROTO_EVENTS_OFF, CA_PROTO_EVENTS_ON,
            CA_PROTO_READ_SYNC, CA_PROTO_ERROR, CA_PROTO_CLEAR_CHANNEL,
            CA_PROTO_RSRV_IS_UP, CA_PROTO_NOT_FOUND, CA_PROTO_READ_NOTIFY,
            CA_PROTO_READ_BUILD, REPEATER_CONFIRM, CA_PROTO_CREATE_CHAN,
            CA_PROTO_WRITE_NOTIFY, CA_PROTO_CLIENT_NAME, CA_PROTO_HOST_NAME,
            CA_PROTO_ACCESS_RIGHTS, CA_PROTO_ECHO, REPEATER_REGISTER,
            CA_PROTO_SIGNAL, CA_PROTO_CREATE_CH_FAIL, CA_PROTO_SERVER_DISCONN };

    private short commandNumber;

    private ChannelAccessCommand(int commandNumber) {
        // We allow an integer parameter to the constructor, so that we do not
        // have to cast the number in each instance declaration. This is okay,
        // because the constructor can only be used within this enum definition.
        this.commandNumber = (short) commandNumber;
    }

    /**
     * Returns the command number. This is the numeric identifier that is sent
     * in the first field of each Channel Access message header, in or to
     * specify the type of the message.
     * 
     * @return command number as used in Channel Access message header.
     */
    public short getCommandNumber() {
        return commandNumber;
    }

    /**
     * Returns the command for a command number. The command number is the
     * numeric identifier stored in the first field of each Channel Access
     * message header.
     * 
     * @param commandNumber
     *            the command number.
     * @return command for the command number.
     * @throws IllegalArgumentException
     *             if <code>commandNumber</code> does not refer to a well-known
     *             command.
     */
    public static ChannelAccessCommand forCommandNumber(short commandNumber) {
        // We treat the command number as an unsigned integer. Probably, we will
        // never have so many commands, that the numbers are negative, but in
        // theory we could have, so this implementation is future proof.
        int commandNumberAsInt = commandNumber & 0xffff;
        if (commandNumberAsInt >= commandNumberToCommand.length) {
            throw new IllegalArgumentException("Unknown command number: "
                    + commandNumberAsInt);
        }
        return commandNumberToCommand[commandNumberAsInt];
    }

}
