/*
 * Copyright 2014-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.apache.commons.lang3.tuple.Pair;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.util.NullTerminatedStringUtil;

/**
 * Representation of a <code>CA_PROTO_ERROR</code> message. This message is sent
 * from a server to a client to signal an error condition.
 * 
 * @author Sebastian Marsching
 * @see ChannelAccessCommand#CA_PROTO_ERROR
 */
public final class ChannelAccessErrorMessage extends ChannelAccessMessage {

    /**
     * Channel CID sent with an error message if the error is not associated
     * with a specific channel.
     */
    public final static int UNKNOWN_CHANNEL_CID = (int) 0xffffffff;

    private int cid;
    private ChannelAccessStatus status;
    private ChannelAccessMessage originalMessage;
    private String messageString;

    /**
     * Constructor for creating a <code>CA_PROTO_ERROR</code> message.
     * 
     * @param channelCID
     *            client-side channel identifier related to the error message.
     *            If the error is not related to a specific channel or the
     *            channel CID is unknown, the value should be set to
     *            {@value #UNKNOWN_CHANNEL_CID}.
     * @param status
     *            Channel Access status code indicating the kind of error.
     * @param originalMessage
     *            message that caused this error. This parameter may be
     *            <code>null</code> if no such message is available. However, if
     *            it is <code>null</code> the <code>messageString</code> must
     *            also be <code>null</code>.
     * @param messageString
     *            textual error message. This parameter may be null if no such
     *            message is available. It must be <code>null</code> if
     *            <code>originalMessage</code> is <code>null</code>.
     * @throws IllegalArgumentException
     *             if <code>originalMessage</code> is <code>null</code> and
     *             <code>messageString</code> is not <code>null</code> and if
     *             <code>originalMessage</code> does not have original headers
     *             (has not been decoded but created through its public
     *             constructor).
     */
    public ChannelAccessErrorMessage(int channelCID,
            ChannelAccessStatus status, ChannelAccessMessage originalMessage,
            String messageString) {
        super(ChannelAccessCommand.CA_PROTO_ERROR);

        if (status == null) {
            throw new NullPointerException("Status must not be null.");
        }
        if (originalMessage == null
                && (messageString != null && !messageString.isEmpty())) {
            throw new IllegalArgumentException(
                    "Message string cannot have a value if original message is null.");
        }
        if (originalMessage != null
                && !originalMessage.hasOriginalMessageHeader()) {
            throw new IllegalArgumentException(
                    "The original message must have its original message headers (must have been decoded from its binary representation).");
        }

        this.cid = channelCID;
        this.status = status;
        this.originalMessage = originalMessage;
        if (messageString == null) {
            messageString = "";
        }
        this.messageString = messageString;
    }

    private ChannelAccessErrorMessage(ChannelAccessMessageHeader messageHeader,
            int channelCID, ChannelAccessStatus status,
            ChannelAccessMessage originalMessage, String messageString) {
        super(ChannelAccessCommand.CA_PROTO_ERROR, messageHeader);

        this.cid = channelCID;
        this.status = status;
        this.originalMessage = originalMessage;
        if (messageString == null) {
            messageString = "";
        }
        this.messageString = messageString;
    }

    /**
     * Returns the channel CID used by the client. This is the identifier that
     * the client uses to identify the channel. This number might be negative or
     * zero. If this error does not affect a specific channel or if the channel
     * CID is not known by the server, the special value
     * {@link #UNKNOWN_CHANNEL_CID} is returned.
     * 
     * @return channel CID assigned to the affected channel by the client or
     *         {@link #UNKNOWN_CHANNEL_CID} if the channel CID is unknown.
     */
    public int getChannelCID() {
        return cid;
    }

    /**
     * Returns the status associated with this error. The status gives more
     * information about the origin of the error. For an error message, the
     * status should never be <code>ECA_NORMAL</code>.
     * 
     * @return status code associated with this error (never <code>null</code>).
     */
    public ChannelAccessStatus getStatus() {
        return status;
    }

    /**
     * <p>
     * Returns the original message that is attached to this error message. This
     * typically is the message sent by the client that caused the error. If no
     * original message was attached to this error message, this method returns
     * <code>null</code>.
     * </p>
     * 
     * <p>
     * <em>Please note:</em> The message returned is incomplete because the
     * offending message's payload is never included in an error message.
     * Therefore, some of the messages original information might not be
     * available and calling methods that need this information might have
     * unexpected results. In particular, the original message cannot be
     * serialized again.
     * </p>
     * 
     * @return original message that caused the error (reconstructed from the
     *         message's header) or <code>null</code> if no original message
     *         header has been attached to this error message.
     */
    public ChannelAccessMessage getOriginalMessage() {
        return originalMessage;
    }

    /**
     * Returns the string describing details about this error. If no such string
     * has been attached to this message, <code>null</code> is returned.
     * 
     * @return string describing this error or <code>null</code>.
     */
    public String getMessageString() {
        return messageString;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        byte[] serializedMessageString = null;
        int serializedMessageStringLength = 0;
        if (messageString != null && !messageString.isEmpty()) {
            Pair<byte[], Integer> serializedMessageStringWithLength = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedBytes(messageString,
                            Integer.MAX_VALUE
                                    - ChannelAccessConstants.LARGE_HEADER_SIZE
                                    * 2,
                            ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT,
                            charset);
            serializedMessageString = serializedMessageStringWithLength
                    .getLeft();
            serializedMessageStringLength = serializedMessageStringWithLength
                    .getRight();
        }
        int payloadSize = 0;
        if (originalMessage != null
                && originalMessage.hasOriginalMessageHeader()) {
            payloadSize += originalMessage.getOriginalMessageHeader()
                    .getHeaderSize();
            if (serializedMessageString != null) {
                payloadSize += serializedMessageStringLength;
            }
        }
        serializeHeader(byteSink, payloadSize, (short) 0, 0, cid,
                status.toStatusCode(), version, maxPayloadSize);
        if (payloadSize > 0) {
            serializeOriginalHeader(byteSink,
                    originalMessage.getOriginalMessageHeader());
            if (serializedMessageString != null) {
                byteSink.putByteArray(serializedMessageString, 0,
                        serializedMessageStringLength);
            }
        }
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        byte[] serializedMessageString = null;
        int serializedMessageStringLength = 0;
        if (messageString != null && !messageString.isEmpty()) {
            Pair<byte[], Integer> serializedMessageStringWithLength = NullTerminatedStringUtil
                    .stringToTruncatedNullTerminatedBytes(messageString,
                            Integer.MAX_VALUE
                                    - ChannelAccessConstants.LARGE_HEADER_SIZE
                                    * 2,
                            ChannelAccessConstants.MESSAGE_SIZE_ALIGNMENT,
                            charset);
            serializedMessageString = serializedMessageStringWithLength
                    .getLeft();
            serializedMessageStringLength = serializedMessageStringWithLength
                    .getRight();
        }
        int payloadSize = 0;
        if (originalMessage != null
                && originalMessage.hasOriginalMessageHeader()) {
            payloadSize += originalMessage.getOriginalMessageHeader()
                    .getHeaderSize();
            if (serializedMessageString != null) {
                payloadSize += serializedMessageStringLength;
            }
        }
        verifyHeader(payloadSize, 0, version, maxPayloadSize);
    }

    private static void serializeOriginalHeader(ByteSink byteSink,
            ChannelAccessMessageHeader originalHeader) {
        // We do not do any checks on the header here: The header is sent as raw
        // data. The reason why we are sending the error message might well-be
        // that header was incorrect, so doing any validation would conflict
        // with the purpose of this message.
        boolean needLargeHeader;
        int payloadSize = originalHeader.getPayloadSize();
        int count = originalHeader.getCount();
        if (payloadSize >= 0xfff || count > 0xffff) {
            needLargeHeader = true;
        } else {
            needLargeHeader = false;
        }
        byteSink.putShort(originalHeader.getCommand());
        if (needLargeHeader) {
            // A large header is signaled using a payload size of 0xffff and a
            // count of 0.
            byteSink.putShort((short) 0xffff);
        } else {
            byteSink.putShort((short) payloadSize);
        }
        byteSink.putShort(originalHeader.getDataType());
        if (needLargeHeader) {
            // A large header is signaled using a payload size of 0xffff and a
            // count of 0.
            byteSink.putShort((short) 0);
        } else {
            byteSink.putShort((short) count);
        }
        byteSink.putInt(originalHeader.getCID());
        byteSink.putInt(originalHeader.getContextSpecific());
        if (needLargeHeader) {
            byteSink.putInt(payloadSize);
            byteSink.putInt(count);
        }
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessErrorMessage other = (ChannelAccessErrorMessage) obj;
        return new EqualsBuilder().append(this.cid, other.cid)
                .append(this.status, other.status)
                .append(this.originalMessage, other.originalMessage)
                .append(this.messageString, other.messageString).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode()).append(cid)
                .append(status).append(originalMessage).append(messageString)
                .toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("cid=");
        sb.append(cid);
        sb.append(", status=");
        sb.append(status != null ? status.getMessage().name() : "null");
        sb.append(", originalMessage=");
        sb.append(originalMessage);
        sb.append(", messageString=");
        if (messageString != null) {
            sb.append('"');
            sb.append(StringEscapeUtils.escapeJava(messageString));
            sb.append('"');
        } else {
            sb.append("null");
        }
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param charset
     *            encoding to use for deserializing the error message (if
     *            present).
     * @param version
     *            protocol version in use. This can determine the meaning of
     *            some header fields.
     * @param isTCP
     *            <code>true</code> if the message to be deserialized has been
     *            received over TCP, <code>false</code> if it has been received
     *            via UDP.
     * @return message constructed from the data in the header and the payload
     *         read from the byte source.
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     */
    protected static ChannelAccessErrorMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            Charset charset, ChannelAccessVersion version, boolean isTCP) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_ERROR
                .getCommandNumber());
        ChannelAccessMessageHeader originalMessageHeader = null;
        ChannelAccessMessage originalMessage = null;
        int payloadSize = messageHeader.getPayloadSize();
        if (payloadSize >= ChannelAccessConstants.STANDARD_HEADER_SIZE) {
            originalMessageHeader = decodeMessageHeader(byteSource,
                    payloadSize >= ChannelAccessConstants.LARGE_HEADER_SIZE);
            try {
                if (isTCP) {
                    originalMessage = ChannelAccessMessageCodec
                            .decodeMessageToUDPServer(originalMessageHeader,
                                    null, true, charset, 0, version);
                } else {
                    originalMessage = ChannelAccessMessageCodec
                            .decodeMessageToTCPServer(originalMessageHeader,
                                    null, true, charset, 0, version);
                }
            } catch (Throwable e) {
                // We do not want an error in the reconstruction of the original
                // message to case the whole error message to be discarded.
                // This means that we cannot throw an exception. We do not want
                // to notify an error handler either, because this would make
                // this deserialize method the only one that uses an error
                // handler. Another option would be to set a flag in the error
                // message itself, indicating that there was a message attached,
                // but it could not be decoded. However, we should then do the
                // same if there is a payload that is too short to contain a
                // message. Anyway, it is not very likely that the calling code
                // would treat such a situation in a special way. Typically, an
                // error message that cannot be correlated with a specific
                // request will simply be logged and otherwise discarded.
            }
        }
        String messageString = null;
        if (originalMessageHeader != null
                && payloadSize > originalMessageHeader.getHeaderSize()) {
            byte[] serializedMessageString = byteSource
                    .getByteArray(payloadSize
                            - originalMessageHeader.getHeaderSize());
            messageString = NullTerminatedStringUtil
                    .nullTerminatedBytesToString(serializedMessageString,
                            charset);
        }

        return new ChannelAccessErrorMessage(messageHeader,
                messageHeader.getCID(),
                ChannelAccessStatus.forStatusCode(messageHeader
                        .getContextSpecific()), originalMessage, messageString);
    }

    private static ChannelAccessMessageHeader decodeMessageHeader(
            ByteSource byteSource, boolean allowLargeHeader) {
        short command = byteSource.getShort();
        short payloadSize = byteSource.getShort();
        short dataType = byteSource.getShort();
        short count = byteSource.getShort();
        int cid = byteSource.getInt();
        int contextSpecific = byteSource.getInt();
        int largePayloadSize;
        int largeCount;
        boolean largeHeader;
        // A payload size of 0xffff signals a large header.
        if (payloadSize == 0xffff && allowLargeHeader) {
            largeHeader = true;
            largePayloadSize = byteSource.getInt();
            largeCount = byteSource.getInt();
        } else {
            largeHeader = false;
            largePayloadSize = payloadSize & 0xffff;
            largeCount = count & 0xffff;
        }
        return new ChannelAccessMessageHeader(command, largePayloadSize,
                dataType, largeCount, cid, contextSpecific, largeHeader);
    }

}
