/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.protocol;

import java.nio.BufferUnderflowException;
import java.nio.charset.Charset;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;

import com.aquenos.epics.jackie.common.io.ByteSink;
import com.aquenos.epics.jackie.common.io.ByteSource;
import com.aquenos.epics.jackie.common.value.ChannelAccessValueType;

/**
 * Representation of a <code>CA_PROTO_READ</code> message that is sent from a
 * client to a server. This message is sent in order to read from a channel.
 * This message should not be used by new clients. They should use the
 * {@link ChannelAccessReadNotifyClientMessage}.
 * 
 * @see ChannelAccessCommand#CA_PROTO_READ
 * @see ChannelAccessReadMessage
 * @see ChannelAccessReadServerMessage
 * 
 * @author Sebastian Marsching
 */
public class ChannelAccessReadClientMessage extends ChannelAccessReadMessage {

    private short dataType;
    private int count;
    private int sid;
    private int operationId;

    /**
     * Creates a <code>CA_PROTO_READ</code> message. This constructor should be
     * used to create a message that is supposed to be sent from a client to a
     * server.
     * 
     * @param dataType
     *            requested data-type for the read operation. The server should
     *            reply with this data type.
     * @param count
     *            the number of value elements to read. Must not be larger than
     *            the channel's native element count. May only be zero for
     *            Channel Access version 4.13 and newer. In this case, a value
     *            of zero indicates to use the channel's native count.
     * @param sid
     *            server-assigned ID identifying the channel.
     * @param operationId
     *            client-assigned ID identifying the read request. This ID will
     *            be sent with the reply so that the client can identify the
     *            request to which the reply belongs.
     */
    public ChannelAccessReadClientMessage(ChannelAccessValueType dataType,
            int count, int sid, int operationId) {
        super();

        this.dataType = dataType.toTypeCode();
        this.count = count;
        this.sid = sid;
        this.operationId = operationId;
    }

    private ChannelAccessReadClientMessage(
            ChannelAccessMessageHeader messageHeader, short dataType,
            int count, int sid, int operationId) {
        super(messageHeader);

        this.dataType = dataType;
        this.count = count;
        this.sid = sid;
        this.operationId = operationId;
    }

    /**
     * Returns the data-type requested by the client. This is the data type that
     * the server should use when sending the reply. Returns <code>null</code>
     * if the specified data-type is unknown to this implementation.
     * 
     * @return data-type for the read request or <code>null</code> if the
     *         numeric identifier cannot be mapped to a data type known by this
     *         implementation.
     */
    public ChannelAccessValueType getDataType() {
        try {
            return ChannelAccessValueType.forTypeNumber(dataType);
        } catch (IllegalArgumentException e) {
            return null;
        }
    }

    /**
     * Returns the number of elements of the value requested by the client. This
     * number should not exceed the native count sent by the server when the
     * channel was connected. A count of zero is only allowed since Channel
     * Access version 4.13. Since this version, this means that the server
     * should use the native count of the channel.
     * 
     * @return number of value elements requested by the client.
     */
    public int getCount() {
        return count;
    }

    /**
     * Returns the channel SID. This is the identifier that the server uses to
     * identify the channel. This number might be negative or zero.
     * 
     * @return channel SID assigned to the channel by the server.
     */
    public int getChannelSID() {
        return sid;
    }

    /**
     * Returns the client-assigned operation ID. The operation ID is used by the
     * client in order to map the reply to this message to the original read
     * request. This number might be negative or zero.
     * 
     * @return client-assigned operation identifier.
     */
    public int getOperationId() {
        return operationId;
    }

    @Override
    protected void serialize(ByteSink byteSink, ChannelAccessVersion version,
            int maxPayloadSize, Charset charset) {
        serializeHeader(byteSink, 0, dataType, count, sid, operationId,
                version, maxPayloadSize);
    }

    @Override
    protected void verify(ChannelAccessVersion version, int maxPayloadSize,
            Charset charset) {
        verifyHeader(0, count, version, maxPayloadSize);
    }

    @Override
    public boolean equals(Object obj) {
        if (!super.equals(obj)) {
            return false;
        }
        // Now we know that the other object is not null and of the right type.
        ChannelAccessReadClientMessage other = (ChannelAccessReadClientMessage) obj;
        return new EqualsBuilder().append(this.dataType, other.dataType)
                .append(this.count, other.count).append(this.sid, other.sid)
                .append(this.operationId, other.operationId).isEquals();
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().appendSuper(super.hashCode())
                .append(dataType).append(count).append(sid).append(operationId)
                .toHashCode();
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append(command.toString());
        sb.append('{');
        sb.append("dataType=");
        try {
            sb.append(ChannelAccessValueType.forTypeNumber(dataType));
        } catch (IllegalArgumentException e) {
            sb.append("UNKNOWN(");
            sb.append(dataType);
            sb.append(')');
        }
        sb.append(", count=");
        sb.append(count);
        sb.append(", sid=");
        sb.append(sid);
        sb.append(", operationId=");
        sb.append(operationId);
        sb.append('}');
        return sb.toString();
    }

    /**
     * Constructs a message from a header and a byte source. This method is only
     * intended for use by {@link ChannelAccessMessageCodec}.
     * 
     * @param messageHeader
     *            CA message header.
     * @param byteSource
     *            byte source for reading the message's payload (if any).
     * @param headerOnly
     *            if <code>true</code> this method will not try to read the
     *            payload but just read the header. This is useful for reading a
     *            message that is sent in the payload of a
     *            {@link ChannelAccessErrorMessage}. Such a message will have
     *            all headers but no payload (even so the payload size might be
     *            non-zero).
     * @return message constructed from the data in the header and the payload
     *         read from the byte source (if <code>headerOnly</code> is
     *         <code>false</code>).
     * @throws BufferUnderflowException
     *             if there is not enough data in the byte source to read the
     *             complete payload.
     */
    protected static ChannelAccessReadClientMessage deserialize(
            ChannelAccessMessageHeader messageHeader, ByteSource byteSource,
            boolean headerOnly) {
        assert (messageHeader.getCommand() == ChannelAccessCommand.CA_PROTO_READ
                .getCommandNumber());
        if (!headerOnly) {
            byteSource.skip(messageHeader.getPayloadSize());
        }
        return new ChannelAccessReadClientMessage(messageHeader,
                messageHeader.getDataType(), messageHeader.getCount(),
                messageHeader.getCID(), messageHeader.getContextSpecific());
    }

}
