/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.util.HashSet;

/**
 * <p>
 * Abstract base implementation of a {@link ListenableFuture}.
 * </p>
 * 
 * <p>
 * This implementation stores the registered listeners in a {@link HashSet} and
 * provides a {@link #notifyCompletionListeners()} method that should be invoked
 * by derived classes.
 * </p>
 * 
 * <p>
 * Derived classes can also override the
 * {@link #notifyCompletionListener(FutureCompletionListener)} method if they
 * want to modify the behavior how listeners are called.
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <V>
 *            The result type returned by the future.
 */
public abstract class AbstractListenableFuture<V> implements
        ListenableFuture<V> {

    /**
     * Synchronization object that this class synchronizes on when modifying the
     * set of registered listeners or triggering the listeners from
     * {@link #notifyCompletionListeners()}.
     */
    protected final Object completionListenerLock = new Object();

    private HashSet<FutureCompletionListener<? super V>> listeners = new HashSet<FutureCompletionListener<? super V>>();
    private boolean notificationDone;

    /**
     * Default constructor. This constructor is supposed to be used by classes
     * that inherit from this class. It initializes the future with an empty set
     * of listeners.
     */
    protected AbstractListenableFuture() {
    }

    /**
     * Notifies all currently registered completion listeners. This method
     * should be called exactly once, after the future's result became
     * available. It delegates calling each listener to
     * {@link #notifyCompletionListener(FutureCompletionListener)}, so that
     * derived classes can take extra precautions (e.g. catch certain exceptions
     * thrown by a listener).
     * 
     * @throws IllegalStateException
     *             if this method has already been called and thus the listeners
     *             have already been notified or if this future has not
     *             completed yet ({@link #isDone()} returns <code>false</code>).
     */
    protected void notifyCompletionListeners() {
        synchronized (completionListenerLock) {
            if (notificationDone) {
                throw new IllegalStateException(
                        "notifyCompletionListeners() has already been called.");
            }
            if (!isDone()) {
                throw new IllegalStateException(
                        "The notifyCompletionListeners() method should only be called if the future has completed (isDone() returns true).");
            }
            // We set the notification done flag before calling the listeners.
            // The flag should be set even if one of the listeners throws an
            // exception.
            notificationDone = true;
            // It certainly is not optimal to call the listeners while we hold
            // the lock. However, if we did not, we could not guarantee that a
            // listener is not notified after having been removed.
            for (FutureCompletionListener<? super V> listener : listeners) {
                notifyCompletionListener(listener);
            }
        }
    }

    /**
     * Notifies a completion listener. This method can be overridden in order to
     * modify how the listener is called. For example, some implementations
     * might want to catch certain exceptions thrown by a listener.
     * 
     * @param listener
     *            the listener that should be notified. Notification should
     *            happen by calling the listeners
     *            {@link FutureCompletionListener#completed(ListenableFuture)
     *            completed()} method, passing a reference to this future
     *            instance.
     */
    protected void notifyCompletionListener(
            FutureCompletionListener<? super V> listener) {
        listener.completed(this);
    }

    @Override
    public boolean addCompletionListener(
            FutureCompletionListener<? super V> listener) {
        // We do not allow adding null listeners, because this does not make any
        // sense and would only lead to problems when trying to notify the
        // listeners.
        if (listener == null) {
            throw new NullPointerException(
                    "The listener parameter must not be null.");
        }
        // We both check whether future has completed and add the listener while
        // holding the lock, so that it only gets added if the future has not
        // completed yet. This ensures that the listener will be called exactly
        // once - either from this method of when the future completes.
        synchronized (completionListenerLock) {
            boolean notRegisteredYet = listeners.add(listener);
            if (notificationDone && notRegisteredYet) {
                notifyCompletionListener(listener);
            }
            return notRegisteredYet;
        }
    }

    @Override
    public boolean removeCompletionListener(
            FutureCompletionListener<? super V> listener) {
        synchronized (completionListenerLock) {
            return listeners.remove(listener);
        }
    }

}
