/*
 * Copyright 2015-2016 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.util;

import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * <p>
 * Simple implementation of a {@link ListenableFuture}.
 * </p>
 * 
 * <p>
 * This implementation provides two methods {@link #setResult(Object)} and
 * {@link #setExecutionException(Throwable)} which can be used to set the result
 * or an exception that occurred while trying to compute the result. Calling
 * these methods will trigger the listeners registered via
 * {@link #addCompletionListener(FutureCompletionListener)} which will be called
 * before the methods return.
 * </p>
 * 
 * @author Sebastian Marsching
 *
 * @param <V>
 *            The result type returned by the future.
 */
public class SimpleListenableFuture<V> extends AbstractListenableFuture<V> {

    private boolean done;
    private Throwable executionException;
    private final Object lock = new Object();
    private V result;

    /**
     * Creates a future. The future has no result and no exception set at the
     * time of creating and is thus not done yet. The
     * {@link #setExecutionException(Throwable)} or {@link #setResult(Object)}
     * has to be called in order to complete the future. After calling either of
     * these methods, the future will be done.
     */
    public SimpleListenableFuture() {
    }

    /**
     * <p>
     * Sets the result provided by this future. The result is returned when the
     * {@link #get()} or {@link #get(long, TimeUnit)} method is called.
     * </p>
     * 
     * <p>
     * Setting the result completes the future and causes the listeners to be
     * notified. After calling this method, {@link #isDone()} is going to return
     * <code>true</code>.
     * </p>
     * 
     * @param result
     *            result of the future. The result is returned by the
     *            <code>get(...)</code> methods. This parameter may be
     *            <code>null</code>.
     * @throws IllegalStateException
     *             if the future has already completed (by calling either
     *             {@link #setExecutionException(Throwable)}) or
     *             {@link #setResult(Object)}.
     * @see #get()
     * @see #get(long, TimeUnit)
     * @see #isDone()
     * @see #setExecutionException(Throwable)
     */
    public void setResult(V result) {
        synchronized (lock) {
            if (done) {
                throw new IllegalStateException(
                        "A result cannot be set after the future completed.");
            }
            this.result = result;
            done = true;
            lock.notifyAll();
            notifyCompletionListeners();
        }
    }

    /**
     * <p>
     * Sets the exception that occurred while executing the job backing this
     * future. This exception is wrapped in an {@link ExecutionException} and
     * thrown when the {@link #get()} or {@link #get(long, TimeUnit)} method is
     * called.
     * </p>
     * 
     * <p>
     * Setting the exception completes the future and causes the listeners to be
     * notified. After calling this method, {@link #isDone()} is going to return
     * <code>true</code>.
     * </p>
     * 
     * @param executionException
     *            the exception that caused the execution to fail. This
     *            exception is wrapped in an {@link ExecutionException} and
     *            thrown by the <code>get(...)</code> methods. This parameter
     *            must not be <code>null</code>.
     * @throws IllegalStateException
     *             if the future has already completed (by calling either
     *             {@link #setExecutionException(Throwable)}) or
     *             {@link #setResult(Object)}.
     * @throws NullPointerException
     *             if <code>executionException</code> is <code>null</code>.
     * @see #get()
     * @see #get(long, TimeUnit)
     * @see #isDone()
     * @see #setResult(Object)
     */
    public void setExecutionException(Throwable executionException) {
        synchronized (lock) {
            if (executionException == null) {
                throw new NullPointerException(
                        "Parameter executionException must not be null.");
            }
            if (done) {
                throw new IllegalStateException(
                        "An exception cannot be set after the future completed.");
            }
            this.executionException = executionException;
            done = true;
            lock.notifyAll();
            notifyCompletionListeners();
        }
    }

    @Override
    public boolean cancel(boolean mayInterruptIfRunning) {
        // We do not support canceling. However, other implementations might.
        // The specification is unclear whether isDone() should always return
        // true after calling this method, even if the job could not be
        // cancelled. According to the documentation of the cancel(...) method
        // it should, but the documentation of the isDone() method indicates
        // that isDone() only returns true if the job has finished, failed, or
        // been cancelled. We implement in this way (isDone() is not affected
        // by cancel(...), because this seems to make more sense.
        return false;
    }

    @Override
    public boolean isCancelled() {
        // As we do not support canceling, this method always returns false.
        return false;
    }

    @Override
    public boolean isDone() {
        synchronized (lock) {
            return done;
        }
    }

    @Override
    public V get() throws InterruptedException, ExecutionException {
        synchronized (lock) {
            if (!done) {
                lock.wait();
                // We should only wake up when the thread is interrupted or the
                // future completed. If the thread is interrupt, wait() throws,
                // so we can assume that we have a result.
                assert done;
            }
            if (executionException != null) {
                throw new ExecutionException(executionException);
            }
            return result;
        }
    }

    @Override
    public V get(long timeout, TimeUnit unit) throws InterruptedException,
            ExecutionException, TimeoutException {
        synchronized (lock) {
            if (!done) {
                unit.timedWait(lock, timeout);
                if (!done) {
                    throw new TimeoutException(
                            "The future did not complete in time.");
                }
            }
            if (executionException != null) {
                throw new ExecutionException(executionException);
            }
            return result;
        }
    }

}
