/*
 * Copyright 2015 aquenos GmbH.
 * All rights reserved.
 * 
 * Please refer to the file NOTICE.txt in the root of the source distribution or
 * in the META-INF directory of the binary archive for licensing information.
 */

package com.aquenos.epics.jackie.common.value;

/**
 * <p>
 * Channel Access values that stores double-precision floating-point numbers and
 * provides information about the alarm status, engineering units, display
 * precision, and alarm and display limits.
 * </p>
 * 
 * <p>
 * This interface should not be implemented by user-defined classes. Code that
 * wants to create an instance implementing this interface should use the
 * methods provided by {@link ChannelAccessValueFactory} instead.
 * </p>
 * 
 * @author Sebatian Marsching
 */
public interface ChannelAccessGraphicsDouble extends ChannelAccessAlarmDouble,
        ChannelAccessFloatingPointGraphicsValue<Double> {

    /**
     * Returns the upper display limit. The upper display limit is the limit
     * that the value does typically not exceed and thus is suitable for
     * choosing the layout of graphic displays (e.g. gauges). A value of zero
     * typically indicates that the limit has not been configured and thus
     * should not be used.
     * 
     * @return upper display limit.
     */
    double getUpperDisplayLimit();

    /**
     * Sets the upper display limit. The upper display limit is the limit that
     * the value does typically not exceed and thus is suitable for choosing the
     * layout of graphic displays (e.g. gauges). A value of zero typically
     * indicates that the limit has not been configured and thus should not be
     * used.
     * 
     * @param upperDisplayLimit
     *            display limit.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setUpperDisplayLimit(double upperDisplayLimit);

    /**
     * Returns the lower display limit. The lower display limit is the limit
     * that the value does typically not undershoot and thus is suitable for
     * choosing the layout of graphic displays (e.g. gauges). A value of zero
     * typically indicates that the limit has not been configured and thus
     * should not be used.
     * 
     * @return lower display limit.
     */
    double getLowerDisplayLimit();

    /**
     * Sets the lower display limit. The lower display limit is the limit that
     * the value does typically not undershoot and thus is suitable for choosing
     * the layout of graphic displays (e.g. gauges). A value of zero typically
     * indicates that the limit has not been configured and thus should not be
     * used.
     * 
     * @param lowerDisplayLimit
     *            lower display limit.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setLowerDisplayLimit(double lowerDisplayLimit);

    /**
     * Returns the upper alarm limit. The upper alarm limit is the limit that,
     * when exceeded, typically raises a major alarm. A value of zero typically
     * indicates that the limit has not been configured and thus should not be
     * used.
     * 
     * @return upper alarm limit.
     */
    double getUpperAlarmLimit();

    /**
     * Sets the upper alarm limit. The upper alarm limit is the limit that, when
     * exceeded, typically raises a major alarm. A value of zero typically
     * indicates that the limit has not been configured and thus should not be
     * used.
     * 
     * @param upperAlarmLimit
     *            upper alarm limit.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setUpperAlarmLimit(double upperAlarmLimit);

    /**
     * Returns the upper warning limit. The upper warning limit is the limit
     * that, when exceeded, typically raises a minor alarm. A value of zero
     * typically indicates that the limit has not been configured and thus
     * should not be used.
     * 
     * @return upper warning limit.
     */
    double getUpperWarningLimit();

    /**
     * Sets the upper warning limit. The upper warning limit is the limit that,
     * when exceeded, typically raises a minor alarm. A value of zero typically
     * indicates that the limit has not been configured and thus should not be
     * used.
     * 
     * @param upperWarningLimit
     *            upper warning limit.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setUpperWarningLimit(double upperWarningLimit);

    /**
     * Returns the lower warning limit. The lower warning limit is the limit
     * that, when undershot, typically raises a minor alarm. A value of zero
     * typically indicates that the limit has not been configured and thus
     * should not be used.
     * 
     * @return lower warning limit.
     */
    double getLowerWarningLimit();

    /**
     * Sets the lower warning limit. The lower warning limit is the limit that,
     * when undershot, typically raises a minor alarm. A value of zero typically
     * indicates that the limit has not been configured and thus should not be
     * used.
     * 
     * @param lowerWarningLimit
     *            lower warning limit.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setLowerWarningLimit(double lowerWarningLimit);

    /**
     * Returns the lower alarm limit. The lower alarm limit is the limit that,
     * when undershot, typically raises a major alarm. A value of zero typically
     * indicates that the limit has not been configured and thus should not be
     * used.
     * 
     * @return lower alarm limit.
     */
    double getLowerAlarmLimit();

    /**
     * Sets the lower alarm limit. The lower alarm limit is the limit that, when
     * undershot, typically raises a major alarm. A value of zero typically
     * indicates that the limit has not been configured and thus should not be
     * used.
     * 
     * @param lowerAlarmLimit
     *            lower alarm limit.
     * @throws UnsupportedOperationException
     *             if this value is read-only ({@link #isReadOnly()} returns
     *             <code>true</code>).
     */
    void setLowerAlarmLimit(double lowerAlarmLimit);

    /**
     * <p>
     * Returns a read-only version of this value. If the value is read-only,
     * this value is returned. Otherwise, a wrapped version of this value, that
     * does not allow modifications is returned.
     * </p>
     * 
     * <p>
     * Please note that the read-only value is not a copy of this value, but
     * just a wrapper around it. This means, that modifications of the wrapped
     * value will result in modifications of the read-only value. Therefore it
     * is bad practice to modify a value that has been used as the base for a
     * read-only value, because code that still has a reference to the read-only
     * value might not expect this change and thus fail unexpectedly.
     * </p>
     * 
     * @return read-only version of this value (never <code>null</code>).
     */
    ChannelAccessGraphicsDouble asReadOnlyValue();

    /**
     * <p>
     * Creates and returns a copy of this object. The object returned is
     * completely independent from this object. This means that modifications to
     * this object will not affect the returned object and vice-versa.
     * </p>
     * 
     * <p>
     * The object returned is guaranteed to be completely identical to this
     * object, including its type. This means that
     * <code>x.clone().equals(x)</code> and
     * <code>x.clone().getClass() == x.getClass()</code> are always
     * <code>true</code>.
     * </p>
     * 
     * <p>
     * The returned value is guaranteed to allow write access, even if this
     * value is read-only. This means that cloning a read-only value is a good
     * way to get a writable copy.
     * </p>
     * 
     * @return a clone of this instance.
     */
    ChannelAccessGraphicsDouble clone();

}
